

#if 0

 ==== RFC2409 ====


8.1.  diffie-hellman-group1-sha1

   The "diffie-hellman-group1-sha1" method specifies the Diffie-Hellman
   key exchange with SHA-1 as HASH, and Oakley Group 2 [RFC2409] (1024-
   bit MODP Group).  This method MUST be supported for interoperability
   as all of the known implementations currently support it.  Note that
   this method is named using the phrase "group1", even though it
   specifies the use of Oakley Group 2.

6.2 Second Oakley Group

   IKE implementations SHOULD support a MODP group with the following
   prime and generator. This group is assigned id 2 (two).

   The prime is 2^1024 - 2^960 - 1 + 2^64 * { [2^894 pi] + 129093 }.
   Its hexadecimal value is

         FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1
         29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD
         EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245
         E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED
         EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE65381
         FFFFFFFF FFFFFFFF

   The generator is 2 (decimal)
#endif


static const unsigned int oakley2[32]=
{
   0xFFFFFFFF, 0xFFFFFFFF, 0xC90FDAA2, 0x2168C234, 0xC4C6628B, 0x80DC1CD1,
   0x29024E08, 0x8A67CC74, 0x020BBEA6, 0x3B139B22, 0x514A0879, 0x8E3404DD,
   0xEF9519B3, 0xCD3A431B, 0x302B0A6D, 0xF25F1437, 0x4FE1356D, 0x6D51C245,
   0xE485B576, 0x625E7EC6, 0xF44C42E9, 0xA637ED6B, 0x0BFF5CB6, 0xF406B7ED,
   0xEE386BFB, 0x5A899FA5, 0xAE9F2411, 0x7C4B1FE6, 0x49286651, 0xECE65381,
   0xFFFFFFFF, 0xFFFFFFFF
};

typedef struct {
  unsigned int len;
  MPINTWORD *data;
} tssh_mpint_t;

static tssh_mpint_t *const_to_mpint(const unsigned int *u32val, int u32len)
{
}


/*** Basic Arithmetic

I have both 64 bit and 32 bit CPUs. Normally, assembly instructions
lets me get the high and low halves of a multiplication result easily.

For instance, if I multiply two 64bit quantities on an Intel CPU,
the high 64 bits of the result are written to an additional register.
By reading that register I could get the whole 128 bit quantity easily.
However, C doesn't let me do that.

Another problem: GCC always uses imul instead of mul. This causes
problems with the upper bits in the result. The best solution is to
Use the mul instruction directly in assembly. 

However, I will do the following for now. On a 32-bit platform, I will
use 32 bit multiplication with 16 bit operands zero extended to 32 bits.

On a 64 bit cpu, I will use 32 bit operands zero extended to 64 bits.

***/

#ifdef TSSH_ARCH32
#define TSSH_WORD  uint16_t
#define TSSH_DWORD uint32_t
#define TSSH_WORD_SIZE 16
#endif

#ifdef TSSH_ARCH64
#define TSSH_WORD uint32_t
#define TSSH_DWORD uint64_t
#define TSSH_WORD_SIZE 32
#endif

#define TSSH_MULWORD(A,B,H,L) do { \
   TSSH_DWORD __R= (TSSH_DWORD) (A) * (TSSH_DWORD) (B); \
   (H)= __R >> TSSH_WORD_SHIFT ; \
   (L)= __R; } while(0)

/***   Addition and subtraction :
   When you add two unsigned N-bit quantities, the result is N+1 bits.
   The carry bit is normally available in the carry flag, but I can't
   get it from C. Again, I will use the same technique as in multiplication.
   The upper half of a TSSH_DWORD will hold the carry information.

   The upper halves will be initialized to zero. */

#define TSSH_ADDWORD(A,B,C,R) do {\
   TSSH_DWORD __R= (TSSH_DWORD) (A) + (TSSH_DWORD) (B); \
   (C)= __R >> TSSH_WORD_SHIFT; \
   (R)= __R; } while(0)

#define TSSH_SUBWORD(A,B,C,R) do {\
   TSSH_DWORD __R= (TSSH_DWORD) (A) - (TSSH_DWORD) (B); \
   (C)= __R >> TSSH_WORD_SHIFT; \
   (R)= __R; } while(0)


/*** 
  Now, I can do big arithmetic using these primitives. A big number
  will be a sequence of TSSH_WORD elements. These will be ordered
  in LSB-first fashion. 

  A multiplication is then:

  mul(A,B):
   R= zero(A.len+B.len+1);
   T= zero(B.len+1);
   for(i=0;i<A.len;i++)
   {
     multiply_word(T, B, A[i]);
     add_shifted(R, T, i);
   }

  multiply_word(T, B, w):
   for(i=0;i<B.len;i++)
     HL= multiply_word(B[i], w)
     HL += Hprev;
     HPrev, T[i]= H : L

OK. Here, it's impossible for HPrev to overflow. 
Let's say we have two integers A and B, with N bits each.
The result of the multiplication R= H:L = A*B is 2N bits. Let's
see how big R can get. 

For maximum R, we need maximum A and B, that is:

  A= B=  111 ...... 111   = 2^N-1
         --------------
           N bits

 R= A*B= (2^N-1)(2^N-1) =  2^2N - 2^N - 2^N + 1
                        = (2^2N+1) - 2^(N+1)   

   Let's write our terms:
                          2N bits
                         -------------
     X =   2^2N+1 =    1 000 .... 0001
     Y =   2^(N+1)=        1 000000000
                             ---------
                               N+1 bits

  If we divide them into 1:N:N parts:

                              N bits    N bits
                              _______  ________ 
                           1: 0.....0: 00....01
        subtract                   10: 00....00
        ----------------------------------------
                           0: 1...110: 00....01

  As you can see, in the maximum case, H is all ones except the
  last bit and L is all zeroes except the last bit. 

  Let's assume that Hprev is also max. ie. Hprev= 11.....1 (N bits)

  If we add these together, we end up with

           1..110:000..01 
  add      0....0:11....1
 ------------------------
           1...11:000...0

 So, there is no overflow, this algorithm will work!

***/

static tssh_mpint_t *tssh_mpint_zero(int len)
{
  tssh_mpint_t *K;
  K= malloc(sizeof(*K));
  K->data= calloc(len,sizeof(K->data[0]));
  K->len= len;
  return K;
}

static tssh_mpint_free(tssh_mpint_t *K)
{
  free(K->data);
  free(K);
}

static void tssh_mpint_mulword
    (tssh_mpint_t *R, tssh_mpint_t *B, TSSH_DWORD a)
{
  TSSH_DWORD M, Hprev;
  TSSH_WORD *bdata, *bend, *rdata;

  bdata= B->data;
  bend= B->data+B->len;
  rdata= R->data;
  Hprev= 0;

  while(bdata!=bend)
  {
    M= a  *  (TSSH_DWORD) *(bdata++)  + Hprev;
    Hprev= M >> TSSH_WORD_SIZE;
    *(rdata++)= M;
  }
  *rdata= Hprev;
}

static void tssh_mpint_add_shifted
    (tssh_mpint_t *R, tssh_mpint_t *B, int shift)
{
  TSSH_DWORD  C, T;
  TSSH_WORD *bdata, *bend, *rdata;

  bdata= B->data;
  bend= B->data+B->len;
  rdata= R->data + shift;
  C= 0;
  while(bdata!=bend)
  {
    T= (TSSH_DWORD) (*bdata++) + (TSSH_DWORD) (*rdata) + C;
    *(rdata++)= T;
    C= T >> TSSH_WORD_SIZE;
  }
  while(C)
  {
    T= (TSSH_DWORD) (*rdata) + C;
    *(rdata++)= T;
    C= T >> TSSH_WORD_SIZE;
  }
}

tssh_mpint_t* tssh_mpint_mul_simple
    (tssh_mpint_t *A,tssh_mpint_t *B)
{
  tssh_mpint_t *R;
  int a,b;
  TSSH_DWORD   M,Hprev;
  TSSH_WORD    L, Hprev, Va, Vb;

  R= tssh_mpint_zero(A->len+B->len+1);
  T= tssh_mpint_zero(B->len+1);

  adata= A->data;
  aend= A->data+A->len;
  aindex= 0;

  while(adata!=aend)
  {
    tssh_mpint_mulword(T, B, *adata);
    tssh_mpint_add_shifted(R, T, aindex);
    adata++;
    aindex++;
  }

  tssh_mpint_free(T);
  return R;
}


/***
   Anyway, my main focus was to do (a*b)%p. not just a*b. The former
   is much more interesting.

   Now, Let's have a,b<p and try to calculate (a*b)%p.

   If I simply calculate a*b, I can never do the % operator on time
   because a*b will be too big with respect to p. Doing a division
   will probably give me a headache. So, I should compute a*b in a 
   (mod p) universe.

   First of all I could do the following:

   Let  a= an a(n-1) a(n-2) ... a(0)

   Let k be the smallest integer where    b*2^k > p

   Also, let c = b*2^k - p

   Now, we could partition a starting from bit k..

   a = an a(n-1) ... a(k)  :  a(k-1) a(k-2) ... a(0)  = AH:AL

   R= (a*b)%p =  (AH:AL*b)%p
              = (AH*2^k*b + AL*b)%p
              = (AH*c + AL*b)%p

  so, we do the following:

  1- determine k and C, if impossible, do a normal multiplication and quit
  2- AL= mask(a,0 .. k-1)  AH= shiftright(a,k)
  3- Add AL*B to accumulator
  4- Set A= AH, B= C, start over

 ***/
