
typedef struct
{
  uint8_t *data;
  uint32_t len;
} tssh_buffer_t;

typedef struct
{
  char *name;
  int type;
  uint32_t perm;
  uint64_t size;
  uint32_t mtime;
} tssh_fileinfo_t;


//
// This struct buffers incoming ssh packets until a complete
// sftp packet is formed.
//
typedef struct
{
  tssh_buffer_t *partial;
  uint32_t psize;
  tssh_buffer_t *next;
} tssh_ftp_msg_queue_t;
 

typedef struct tssh_write_buffer
{
  uint32_t size;
  uint32_t len;
  uint8_t *data;
  struct tssh_write_buffer *next;
} tssh_write_buffer_t;

typedef struct
{
  void **data;
  uint32_t *len;
  uint32_t eltsize, size;
} tssh_array_t;

typedef struct
{
  char *name;
  int value;
} tssh_algoname_t;

typedef struct
{
  int cmd;
  void *p1, *p2, *p3;
  uint32_t i1;
} tssh_cmd_t;


typedef struct
{
  int len, head, size;
  tssh_cmd_t **data;
  pthread_mutex_t mutex;
  pthread_cond_t have_space, have_data;
} tssh_queue_t;

typedef struct tssh_ftop
{
  int cmd;
  char *fn; char *oldfn;
  char *localfn;
  uint8_t *data;
  uint32_t size;
  struct tssh_ftop *next;
  uint8_t *udata;
} tssh_ftop_t;


typedef struct tssh_hash tssh_hash_t;
typedef struct tssh_crypto tssh_crypto_t;
typedef struct tssh_hmac tssh_hmac_t;

struct tssh_hash
{
  int type;
  uint32_t len;
  uint8_t *data;
  union
  {
   SHA1_CTX sha1;
  } C;
};


struct tssh_crypto
{
  int type;
  union
  {
    struct AES256_ctx aes256ctr;
  } C;
};

struct tssh_crypto_def
{
  char *name;
  int type;
  int block_len;
  int key_len;
};

struct tssh_hmac
{
  int type;
  uint8_t *data;
  int len;
  uint8_t *key_ipad;
  uint8_t *key_opad;
  int keylen;
  union
  {
    SHA1_CTX sha1;
  } C;
};

struct tssh_hmac_def
{
  char *name;
  int key_len;
  int type;
};

// hash functions
enum { TSSH_SHA1= 1 };

// encryption algorithms
enum { TSSH_AES256CTR=1 };

// kex algorithms
enum { TSSH_KEX_DHG14_SHA1=1 };

// server key algorithms
enum { TSSH_SERVERKEY_SSH_RSA=1 };

// compression algorithms
enum { TSSH_COMPRESS_NONE=1 };


#define SSH_MSG_DISCONNECT                        1
#define SSH_MSG_IGNORE                            2
#define SSH_MSG_UNIMPLEMENTED                     3
#define SSH_MSG_DEBUG                             4
#define SSH_MSG_SERVICE_REQUEST                   5
#define SSH_MSG_SERVICE_ACCEPT                    6
#define SSH_MSG_EXT_INFO                          7
#define SSH_MSG_KEXINIT                          20
#define SSH_MSG_NEWKEYS                          21
#define SSH_MSG_KEXDH_INIT                       30
#define SSH_MSG_KEXDH_REPLY                      31
#define SSH_MSG_USERAUTH_REQUEST                 50
#define SSH_MSG_USERAUTH_FAILURE                 51
#define SSH_MSG_USERAUTH_SUCCESS                 52
#define SSH_MSG_USERAUTH_BANNER                  53
#define SSH_MSG_USERAUTH_PASSWD_CHANGEREQ        60
#define SSH_MSG_GLOBAL_REQUEST                   80
#define SSH_MSG_REQUEST_SUCCESS                  81
#define SSH_MSG_REQUEST_FAILURE                  82
#define SSH_MSG_CHANNEL_OPEN                     90
#define SSH_MSG_CHANNEL_OPEN_CONFIRMATION        91
#define SSH_MSG_CHANNEL_OPEN_FAILURE             92
#define SSH_MSG_CHANNEL_WINDOW_ADJUST            93
#define SSH_MSG_CHANNEL_DATA                     94
#define SSH_MSG_CHANNEL_EXTENDED_DATA            95
#define    SSH_EXTENDED_DATA_STDERR               1
#define SSH_MSG_CHANNEL_EOF                      96
#define SSH_MSG_CHANNEL_CLOSE                    97
#define SSH_MSG_CHANNEL_REQUEST                  98
#define SSH_MSG_CHANNEL_SUCCESS                  99
#define SSH_MSG_CHANNEL_FAILURE                 100


#define SSH_FXP_INIT                              1
#define SSH_FXP_VERSION                           2
#define SSH_FXP_OPEN                              3
#define SSH_FXP_CLOSE                             4
#define SSH_FXP_READ                              5
#define SSH_FXP_WRITE                             6
#define SSH_FXP_LSTAT                             7
#define SSH_FXP_FSTAT                             8
#define SSH_FXP_SETSTAT                           9
#define SSH_FXP_FSETSTAT                         10
#define SSH_FXP_OPENDIR                          11
#define SSH_FXP_READDIR                          12
#define SSH_FXP_REMOVE                           13
#define SSH_FXP_MKDIR                            14
#define SSH_FXP_RMDIR                            15
#define SSH_FXP_REALPATH                         16
#define SSH_FXP_STAT                             17
#define SSH_FXP_RENAME                           18
#define SSH_FXP_READLINK                         19
#define SSH_FXP_LINK                             21
#define SSH_FXP_BLOCK                            22
#define SSH_FXP_UNBLOCK                          23
#define SSH_FXP_STATUS                          101
#define SSH_FXP_HANDLE                          102
#define SSH_FXP_DATA                            103
#define SSH_FXP_NAME                            104
#define SSH_FXP_ATTRS                           105
#define SSH_FXP_EXTENDED                        200
#define SSH_FXP_EXTENDED_REPLY                  201


#define SSH_FXF_READ                     0x00000001
#define SSH_FXF_WRITE                    0x00000002
#define SSH_FXF_APPEND                   0x00000004
#define SSH_FXF_CREAT                    0x00000008
#define SSH_FXF_TRUNC                    0x00000010
#define SSH_FXF_EXCL                     0x00000020
					  

typedef struct
{
  uint32_t len;
  uint8_t *data;
} tssh_obj_t;


typedef struct tssh tssh_t;

typedef void (*tssh_command_t)(tssh_t *);


typedef struct
{
  buffer_t *buffer;
  int len;
} tssh_opak_t;

typedef struct
{
  tssh_crypto_t crypto;
  tssh_hmac_t hmac;
  struct tssh_crypto_def *cryptodef;
} tssh_encoder_t;

typedef struct 
{
  char cookie[16];
  tssh_mpint_t *X, *P;
  tssh_numbers_t *P_table;
  tssh_obj_t *msg_init_c2s,
             *msg_init_s2c,
             *host_key,
             *signature,
             *E, *F, *K, *H;
  tssh_encoder_t *encoder_i;

  struct tssh_hmac_def *macdef_c2s, *macdef_s2c;
  struct tssh_crypto_def *cryptodef_c2s, *cryptodef_s2c;

  int kexalgo, serverkeyalgo, compalgo_c2s, compalgo_s2c;

  int state;
} tssh_kex_t;

enum { TSSH_CHANNEL_SHELL=1, TSSH_CHANNEL_PORT_FORWARD, TSSH_CHANNEL_FTP };

typedef struct tssh_global_request
{
  int type; void *pval1; uint32_t ival1;
  struct tssh_global_request *next;
} tssh_global_request_t;

typedef struct
{
  int type;
  int state;
  uint32_t local_number, remote_number;
  uint32_t window_size_s2c, window_size_c2s;
  uint32_t max_packet_size;
  tssh_t *transport;

  // for shell channels
  char **env;
  int sent_env;

  pthread_mutex_t pend_mutex;
  pthread_cond_t pend_cv;
  tssh_write_buffer_t *pend_first, *pend_last;
  uint32_t pend_size;
  uint32_t pend_max;
  int close_sent;

  tssh_ftp_msg_queue_t ftp_mq;
} tssh_channel_t;

typedef struct
{
  int type;
  tssh_channel_t *channel;
  uint32_t connected_port;
  char *connected_address;
  uint32_t originator_port;
  char *originator_address;
  tssh_buffer_t *buffer;
  uint32_t port;
  union
  {
    int reason;
    char *str;
    uint32_t ival;
  } D;
  uint32_t n_fileinfo;
  tssh_fileinfo_t *fileinfo;
  char *path;
  int opcode;
  tssh_ftop_t *ftp_op;
} tssh_event_t;

enum {
  TSSH_EVENT_LOG= 1,
  TSSH_EVENT_CONNECT_BANNER,
  TSSH_EVENT_CONNECT_BANNER_END,
  TSSH_EVENT_PASSWORD_BANNER,
  TSSH_EVENT_CONNECT_SUCCESS,
  TSSH_EVENT_CONNECT_FAILURE,
  TSSH_EVENT_SHELL_OK,
  TSSH_EVENT_SHELL_FAILURE,
  TSSH_EVENT_CHANNEL_DATA,
  TSSH_EVENT_CHANNEL_EXTENDED_DATA,
  TSSH_EVENT_CHANNEL_WRITE_OK,
  TSSH_EVENT_CHANNEL_CLOSED,
  TSSH_EVENT_PORT_FORWARD_SUCCESS,
  TSSH_EVENT_PORT_FORWARD_FAILURE,
  TSSH_EVENT_PORT_FORWARD_CONNECT,

  TSSH_EVENT_FTP_DISCONNECTED,
  TSSH_EVENT_FTP_SUCCESS,
  TSSH_EVENT_FTP_FAILURE,

  TSSH_EVENT_FTP_CONNECT_SUCCESS,
  TSSH_EVENT_FTP_CONNECT_FAILURE,
};

struct tssh
{
   char *host, *alias;
   int port;
   char *user, *password;
  int fd;

  tssh_obj_t *o_remote_id, *o_self_id;
  tssh_obj_t *session_id;

  tssh_kex_t *kex;

  uint32_t oQ;

  int auth, auth_state;
  int have_keys;

  pthread_t main_thread, writer_thread, reader_thread;
  tssh_queue_t  main_queue, writer_queue, reader_queue,
                iface_queue;  // events from the system to user thread

  int iface_fds[2];

  tssh_event_t *last_event;

  int avail_channel_no;
  int nchannels, schannels;
  tssh_channel_t **channels;

  tssh_global_request_t *gr_first, *gr_last;  

  tssh_array_t a_fwreq;
  uint32_t nfwreq;
  uint32_t *fwreq;

  tssh_channel_t *ftp_channel;
  coth_t *ftp_ctx, *main_ctx;
  int ftp_reqid;
  tssh_ftop_t *ftop_first, *ftop_last, *ftop_current; 
};


enum {
  TSSH_CMD_WRITE=1,
  TSSH_CMD_ERROR,
  TSSH_CMD_SETENCODER,
  TSSH_CMD_QUIT,
  TSSH_CMD_START_READER,
  TSSH_CMD_NEWPACKET,
  TSSH_CMD_SHELL,
  TSSH_CMD_CHANNEL_WRITE, // from user thread to main thread
                         // an also from main thread to writer thread
  TSSH_CMD_CHANNEL_WRITE_OK,  // from writer thread to main thread
  TSSH_CMD_PORT_FORWARD,

  //
  // FTP Commands, part 1 : main thread to ftp thread.
  // 
  TSSH_CMD_FTP_DISCONNECT,
  TSSH_CMD_FTP_DATA,
  TSSH_CMD_FTP_CHANNEL_CREATE_OK,
  TSSH_CMD_FTP_CHANNEL_CREATE_FAIL,
  TSSH_CMD_FTP_CHANNEL_REQUEST_OK,
  TSSH_CMD_FTP_CHANNEL_REQUEST_FAIL,

  // FTP Commands, part 2: user thread to main thread
  TSSH_CMD_FTP_CONNECT,
  TSSH_CMD_QUEUE_FTOP,
  // TSSH_CMD_FTP_DISCONNECT   This originates from user to main,
  //                           but then is forwarded to the ftp thread

  // FTP Commands, part 3: ftp thread to main thread
  TSSH_CMD_DEALLOC_COTH,
};

enum {
  TSSH_GR_PORT_FORWARD= 1,
};


#define SSH_OPEN_ADMINISTRATIVELY_PROHIBITED          1
#define SSH_OPEN_CONNECT_FAILED                       2
#define SSH_OPEN_UNKNOWN_CHANNEL_TYPE                 3
#define SSH_OPEN_RESOURCE_SHORTAGE                    4 

#define SSH_FX_OK                                     0
#define SSH_FX_EOF                                    1
#define SSH_FX_NO_SUCH_FILE                           2
#define SSH_FX_PERMISSION_DENIED                      3
#define SSH_FX_FAILURE                                4
#define SSH_FX_BAD_MESSAGE                            5
#define SSH_FX_NO_CONNECTION                          6
#define SSH_FX_CONNECTION_LOST                        7
#define SSH_FX_OP_UNSUPPORTED                         8      

#define SSH_FILEXFER_ATTR_SIZE                        0x00000001
#define SSH_FILEXFER_ATTR_UIDGID                      0x00000002
#define SSH_FILEXFER_ATTR_PERMISSIONS                 0x00000004
#define SSH_FILEXFER_ATTR_ACMODTIME                   0x00000008
#define SSH_FILEXFER_ATTR_EXTENDED                    0x80000000      

#define TSSH_FTOP_LIST                               1
#define TSSH_FTOP_UPLOADMEM                          2
#define TSSH_FTOP_UPLOADFILE                         3
#define TSSH_FTOP_DOWNLOADMEM                        4
#define TSSH_FTOP_DOWNLOADFILE                       5
#define TSSH_FTOP_RMFILE                             6
#define TSSH_FTOP_RMDIR                              7
#define TSSH_FTOP_RENAME                             8
#define TSSH_FTOP_QUERY                              9
#define TSSH_FTOP_MKDIR                              10



#define TSSH_OK                                       0
#define TSSH_FAIL                                     1
#define TSSH_FAIL_HARD                                2
#define TSSH_EOF                                      3

