
static unsigned char tssh_get_byte(unsigned char *K)
{
  return K[0];
}

static uint16_t tssh_get_uint16(unsigned char *K)
{
  return (K[0]<<8) | K[1];
}

static uint32_t tssh_get_uint32(unsigned char *K)
{
  return (K[0]<<24) | (K[1]<<16) | (K[2]<<8) | K[3];
}

static uint64_t tssh_get_uint64(unsigned char *K)
{
  uint64_t V;
  int i;
  V= 0;
  for(i=0;i<8;i++)
  {
    V= (V<<8) | K[i];
  }
  return V;
}

static void tssh_put_uint64(unsigned char *K, uint64_t V)
{
  int i;
  for(i=7;i>=0;i--)
  {
    K[i]= V & 0xff;
    V>>= 8;
  }
}

static uint8_t tssh_get_uint8(unsigned char *K)
{
  return K[0];
}

static void tssh_put_uint8(unsigned char  *K, uint8_t V)
{
  K[0]= V;
}

static void tssh_put_uint16(unsigned char *D, uint16_t V)
{
  D[0]= V>>8;
  D[1]= V;
}

static void tssh_put_uint32(unsigned char *D, uint32_t V)
{
  D[0]= V>>24;
  D[1]= V>>16;
  D[2]= V>>8;
  D[3]= V;
}

static tssh_obj_t *tssh_obj_new(uint8_t *data,uint32_t len);

static tssh_obj_t *tssh_mpint_to_obj(tssh_mpint_t *N)
{
  uint8_t *data;
  uint32_t len;
  tssh_mpint_to_rfc_mpint(N, &len, &data);
  return tssh_obj_new(data,len);
}

static uint32_t tssh_min_uint32(uint32_t a, uint32_t b)
{
  return a<b? a:b;
}

static void tssh_array_grow(tssh_array_t *A)
{
  if (A->len[0]< A->size) return ;
  A->size+= A->size/2 + 10;
  A->data[0]= realloc(A->data[0], A->eltsize*A->size);
}

static void tssh_array_init
   (tssh_array_t *A, void **data, uint32_t *len,
    uint32_t eltsize, uint32_t initsize)
{
   A->len= len;
   A->data= data;
   A->eltsize= eltsize;
   A->size= initsize;
   A->data[0]= malloc( eltsize*initsize );
   A->len[0]= 0;
}

static void tssh_checkfree(void *a)
{
  if (a) free(a);
}
