
#ifdef TSSH_NO_DEBUG

#define NOP do { } while(0)

#define tssh_print_message(X,Y) NOP
#define tssh_dump_bytes(X,Y,Z) NOP
#define tssh_dump_obj(X,Y) NOP
#define tssh_mpint_print(X,Y) NOP
#define tssh_debug_hex(X,Y,Z,...) NOP

#else

static int dchr(char x) {
  if (x>=' ' && x<127) return x;
  return '.';
}

static void tssh_dump_bytes(char *msg, unsigned char *mem, int len)
{
  int i,j;
  printf("--- BEGIN %s : %d bytes ---\n", msg, len);
  for(i=0;i<len;i+=16)
  {
    printf("%04d:",i);
    for(j=0;j<16 && j+i<len;j++) printf(" %02x", mem[i+j]);
    for(;j<16;j++) printf("   ");
    printf("  ");
    for(j=0;j<16 && j+i<len;j++) printf("%c", dchr(mem[i+j]));
    for(;j<16;j++) printf(".");
    printf("\n");
  }
  printf("--- END %s ---\n", msg);
}

static void tssh_debug_hex(unsigned char *mem, int len,unsigned char *msg,...)
{
  int i,j;
  va_list ap;
  printf("--- BEGIN: ");
  va_start(ap,msg);
  vfprintf(stdout, msg, ap);
  va_end(ap);
  printf(" (%d bytes) ---\n",len);
  for(i=0;i<len;i+=16)
  {
    printf("%04d:",i);
    for(j=0;j<16 && j+i<len;j++) printf(" %02x", mem[i+j]);
    for(;j<16;j++) printf("   ");
    printf("  ");
    for(j=0;j<16 && j+i<len;j++) printf("%c", dchr(mem[i+j]));
    for(;j<16;j++) printf(".");
    printf("\n");
  }
  printf("--- END: ");
  va_start(ap,msg);
  vfprintf(stdout, msg, ap);
  va_end(ap);
  printf("---\n");
}

static void tssh_dump_obj(char *msg, tssh_obj_t *obj)
{
  tssh_dump_bytes(msg, obj->data, obj->len);
}

static void tssh_print_bytes(unsigned char **Rmsg, int *Rlen,int len)
{
  int i;
  if (*Rlen<=0) return ;
  for(i=0;i<len && i<*Rlen;i++) printf("%02x ", (*Rmsg)[i]);
  printf("\n");
  *Rmsg += i;
  *Rlen -= i;
}

static void tssh_print_string_hex(unsigned char **Rmsg, int *Rlen)
{
  unsigned char *msg;
  uint32_t SL;
  int len;
  int i;
  msg= *Rmsg;
  len= *Rlen;

  SL= tssh_get_uint32(msg); msg+= 4;
  for(i=0;i<SL;i++) printf("%02x ", msg[i]);
  printf("\n");

  *Rmsg= msg + SL;
  *Rlen= len - 4- SL;
}

static void tssh_print_namelist(unsigned char **Rmsg, int *Rlen)
{
  unsigned char *K; int len;
  int sz;
  K= *Rmsg;
  len= *Rlen;

  sz= tssh_get_uint32(K);
  fwrite(K+4, 1, sz, stdout);
  printf("\n");
  *Rmsg += sz+4;
  *Rlen -= sz+4;
}

typedef struct
{
  int msg;
  char *name;
  char fmt;
  int ivalue;
} tssh_debug_msg_format_t;

// For each message, the first one is the message name
// if this entry has 'x', only the message name will
// be printed, in order to reduce the amount of message
// log printed. 
static tssh_debug_msg_format_t tssh_message_formats[]=
{
  { SSH_MSG_KEXINIT,      "kex_init" ,'x' },
  { 0,                    "Cookie", 'B', 16 },
  { 0,                    "Kex-Algorithms", 'n' },
  { 0,                    "Server Host Key Algorithms", 'n',},
  { 0,                    "Encryption c2c", 'n', },
  { 0,                    "Encryption s2c", 'n', },
  { 0,                    "MAC c2s", 'n', },
  { 0,                    "MAC s2c", 'n', },
  { 0,                    "Compression c2c", 'n', },
  { 0,                    "Compression s2c", 'n', },
  { 0,                    "Languages c2s", 'n', },
  { 0,                    "Languages s2c", 'n', },
  { 0,                    "First Kex Packet Follows", 'b', }, 
  { 0,                    "Reserved", 'u' },
  { -1,                   NULL, 0 },
  { SSH_MSG_KEXDH_REPLY,  "kexdh_reply" , 'x' },
  { 0,                    "Host Key", 'h' },
  { 0,                    "F", 'h' },
  { 0,                    "Signature", 'h' },
  { -1,                   NULL, 0 },
  { SSH_MSG_NEWKEYS,      "newkeys" },
  { -1,                   NULL, 0 },
  { SSH_MSG_USERAUTH_SUCCESS, "userauth_success" },
  { -1,                   NULL, 0 },
  { SSH_MSG_USERAUTH_FAILURE, "userauth_failure" },
  { 0,                    "authmethods", 'n', },
  { 0,                    "partial success", 'b', },
  { -1,                   NULL, 0 },
  { SSH_MSG_SERVICE_ACCEPT, "service_accept" },
  { 0                     , "service", 'n' },
  { -1,                   NULL, 0 },
  { SSH_MSG_IGNORE,         "ignore", },
  { -1,                   NULL, 0 },
  { SSH_MSG_DEBUG,         "debug", },
  { 0,                     "always_display", 'b', },
  { 0,                     "msg", 'n' },
  { 0,                     "lang", 'n' },
  { -1,                   NULL, 0 },
  { SSH_MSG_GLOBAL_REQUEST, "global_request", },
  { 0,                    "request_name", 'n' },
  { 0,                    "want_reply", 'b', },
  { -1,                   NULL, 0 },
  
  { -2,                   NULL, 0 },
};

/***
messages to go
#define SSH_MSG_DISCONNECT                        1
#define SSH_MSG_UNIMPLEMENTED                     3
#define SSH_MSG_EXT_INFO                          7
#define SSH_MSG_USERAUTH_PASSWD_CHANGEREQ        60
#define SSH_MSG_GLOBAL_REQUEST                   80
#define SSH_MSG_REQUEST_SUCCESS                  81
#define SSH_MSG_REQUEST_FAILURE                  82
#define SSH_MSG_CHANNEL_OPEN                     90
#define SSH_MSG_CHANNEL_OPEN_CONFIRMATION        91
#define SSH_MSG_CHANNEL_OPEN_FAILURE             92
#define SSH_MSG_CHANNEL_WINDOW_ADJUST            93
#define SSH_MSG_CHANNEL_DATA                     94
#define SSH_MSG_CHANNEL_EXTENDED_DATA            95
#define SSH_MSG_CHANNEL_EOF                      96
#define SSH_MSG_CHANNEL_CLOSE                    97
#define SSH_MSG_CHANNEL_REQUEST                  98
#define SSH_MSG_CHANNEL_SUCCESS                  99
#define SSH_MSG_CHANNEL_FAILURE                 100

  -- we send the following messages, so no need to print them really.
#define SSH_MSG_SERVICE_REQUEST                   5
#define SSH_MSG_KEXDH_INIT                       30
#define SSH_MSG_USERAUTH_REQUEST                 50
****/


static void tssh_mpint_print(char *msg,tssh_mpint_t *T)
{
  int i,j;
  char *hexchars= "0123456789ABCDEF";
  TSSH_WORD W,V;
  printf("%s ",msg);
  for(i=T->len-1;i>=0;i--)
  {
    W= T->data[i];
    for(j=TSSH_WORD_SIZE/4-1;j>=0;j--)
    {
      V= W >> (j*4);
      V &= 0xf;
      printf("%c", hexchars[V]);
    }
    printf(" ");
  }
  printf("\n");
}

static void tssh_print_message(uint8_t *msg, int len)
{
  int i;
  int cmd;
  tssh_debug_msg_format_t *F;
  F= tssh_message_formats;
  cmd= msg[0];
  msg++; len--;
  for(i=0;F[i].msg!=-2;i++) if (F[i].msg==cmd) goto found;
  tssh_debug_hex(msg,len, "Message %d", cmd);
  return ;
found:
  printf("Message(%s)\n", F[i].name);
  if (F[i].fmt=='x') return ; // just skip the details here
  i++;
  for(;F[i].msg!=-1 && len>0;i++)
  switch(F[i].fmt)
  {
  case 'n': printf("%s: ", F[i].name); 
            tssh_print_namelist(&msg,&len); 
            break;
  case 'h': printf("%s: ", F[i].name);
            tssh_print_string_hex(&msg, &len); 
            break;
  case 'b': printf("%s: %d\n", F[i].name, msg[0]);
            msg++; len--;
            break;
  case 'B': printf("%s: ", F[i].name);
            tssh_print_bytes(&msg, &len, F[i].ivalue);
            break;
  case 'u': printf("%s: %x\n", F[i].name, tssh_get_uint32(msg));
            msg+= 4;
            len-= 4;
            break;
  }
  if (len>0)
    tssh_debug_hex(msg, len, "Remaining %d bytes", len);
}
#endif

static int tssh_logv(const char *fmt, va_list ap, char *buf)
{
  int L,aL;
  char *str;
  tssh_obj_t *obj;
  char ibuf[20],C;

  L= 0;
again:
  while(*fmt)
  {
    if (fmt[0]=='%') break;
    if (buf) *(buf++)= *fmt;
    fmt++;
    L++;
  }
  if (!*fmt) goto finish;
  fmt++; // skip the %
  C= *(fmt++); // skip the format char
  switch(C)
  {
  case 0: if (buf) *(buf++)= '%'; L++; goto finish;
  case 'd': sprintf(ibuf, "%d", va_arg(ap,int)); 
            str= ibuf;
            aL= strlen(ibuf);
            goto put_str;
  case 's': str= va_arg(ap,char*);
            aL= strlen(str);
            goto put_str;
  case 'S': obj= va_arg(ap, tssh_obj_t*);
            str= obj->data+4;
            aL= obj->len-4;
            goto put_str;
  default: if (buf) *(buf++)= C; L++; goto again;
  }
finish:
  if (buf) *(buf++)= 0;
  return L;

put_str:
  if (buf) { memcpy(buf, str, aL); buf+= aL; }
  L+= aL;
  goto again;
}

#ifdef TSSH_NO_DEBUG
static void tssh_log(tssh_t *T, const char *fmt, ...)
{
  va_list ap;
  int n;
  char *buf;
  tssh_event_t *ev;

  va_start(ap,fmt);
  n= tssh_logv(fmt, ap, NULL);
  va_end(ap);

  buf= malloc(n+1);

  va_start(ap,fmt);
  tssh_logv(fmt, ap, buf);
  va_end(ap);

  ev= tssh_event_new(TSSH_EVENT_LOG);
  ev->D.str= buf;
  tssh_report(T, ev);
}
#else
static void tssh_log(tssh_t *T, const char *fmt, ...)
{
  va_list ap;
  int n;
  char *buf;

  va_start(ap,fmt);
  n= tssh_logv(fmt, ap, NULL);
  va_end(ap);

  buf= malloc(n+1);

  va_start(ap,fmt);
  tssh_logv(fmt, ap, buf);
  va_end(ap);

  fwrite(buf, 1, strlen(buf), stdout);
  free(buf);
}
#endif
