
static tssh_obj_t *tssh_obj_new(uint8_t *data, uint32_t len)
{
  tssh_obj_t *T;
  T= malloc(sizeof(*T));
  T->data= data;
  T->len= len;
  return T;
}

static tssh_obj_t *tssh_obj_copy(uint8_t *data, uint32_t len)
{
  uint8_t *ndata;
  ndata= malloc(len);
  memcpy(ndata,data,len);
  return tssh_obj_new(ndata,len);
}

static tssh_obj_t *tssh_objdup(tssh_obj_t *O)
{
  return tssh_obj_copy(O->data, O->len);
}

static tssh_obj_t *tssh_str_to_rfc(char *str)
{
  uint8_t *data;
  uint32_t L;
  L= strlen(str);
  data= malloc(L+4);
  tssh_put_uint32(data, L);
  memcpy(data+4, str, L);
  return tssh_obj_new(data, L+4);
}


static void tssh_obj_free_deep(tssh_obj_t* obj)
{
  free(obj->data);
  free(obj);
}

static void tssh_obj_mark(tssh_obj_t *obj, tssh_obj_t *mark)
{
  mark->data= obj->data;
  mark->len= obj->len;
}

static void tssh_obj_restore(tssh_obj_t *obj, tssh_obj_t *mark)
{
  obj->data= mark->data;
  obj->len= mark->len;
}

static int tssh_xstr_compare(tssh_obj_t *obj, char *str)
{
  int L;
  L= strlen(str);
  if (obj->len!=L+4) return 1;
  return memcmp(obj->data+4, str, L);
}

static char *tssh_xstr_to_str(tssh_obj_t *obj)
{
  char *data;
  data= malloc(obj->len-4+1);
  memcpy(data, obj->data+4, obj->len-4);
  data[obj->len-4]= 0;
  return data;
}

static tssh_buffer_t *tssh_buffer_new(uint32_t size)
{
  tssh_buffer_t *Q;
  Q= malloc(sizeof(*Q)+size);
  Q->len= 0;
  Q->data= (char*) Q + sizeof(*Q);
  return Q;
}



static int tssh_parsecheck(tssh_buffer_t *P, int len)
{
  if (P->len<len) return 1;
  return 0;
}

static int tssh_parse_uint32(tssh_buffer_t *P, uint32_t *R)
{
  if (tssh_parsecheck(P, 4)) return 1;
  *R= tssh_get_uint32(P->data);
  P->data+= 4;
  P->len-= 4;
  return 0;
}

static int tssh_parse_uint64(tssh_buffer_t *P, uint64_t *R)
{
  if (tssh_parsecheck(P, 8)) return 1;
  *R= tssh_get_uint64(P->data);
  P->data+= 8;
  P->len-= 8;
  return 0;
}

static int tssh_parse_byte(tssh_buffer_t *P, uint8_t *R)
{
  if (tssh_parsecheck(P, 1)) return 1;
  *R= *P->data;
  P->data+= 1;
  P->len-= 1;
  return 0;
}

static int tssh_parse_bytes(tssh_buffer_t *P, uint8_t **R, int len)
{
  if (tssh_parsecheck(P, len)) return 1;
  *R= P->data;
  P->data+= len;
  P->len-= len;
  return 0;
}

static int tssh_parse_bool(tssh_buffer_t *P, uint8_t *R)
{
  return tssh_parse_byte(P, R);
}

static int tssh_parse_string(tssh_buffer_t*P, tssh_obj_t *O)
{
  uint32_t L;
  uint8_t *OP;
  O->data= P->data;
  if (tssh_parse_uint32(P, &L)) return 1;
  if (tssh_parsecheck(P, L)) return 1;
  P->data+= L;
  P->len-= L;
  O->len= L+4;
  return 0;
}

static int tssh_parse_namelist(tssh_buffer_t*P, tssh_obj_t *O)
{
  return tssh_parse_string(P,O);
}

static int tssh_parse_mpint(tssh_buffer_t*P, tssh_obj_t *O)
{
  return tssh_parse_string(P,O);
}

static tssh_obj_t* tssh_parse_string_copy(tssh_buffer_t* P)
{
  tssh_obj_t O;
  if (tssh_parse_string(P,&O)) return NULL;
  return tssh_objdup(&O);
}

static tssh_obj_t* tssh_parse_mpint_copy(tssh_buffer_t* P)
{
  return tssh_parse_string_copy(P);
}

static void tssh_namelist_skiplen(tssh_obj_t *L)
{
  L->data+= 4;
  L->len-= 4;
}

static int tssh_namelist_get(tssh_obj_t *L, tssh_obj_t *R)
{
  int f,i;
  if (L->len==0) return 0;
  R->data= L->data;
  f= 0;
  for(i=0;i<L->len;i++)
   if (L->data[i]==',') { f= 1; break; }
  R->len= i;
  L->data += i+f;
  L->len -= i+f;
  return 1;
}



static tssh_obj_t *tssh_obj_copy_str
   (uint8_t *end, uint8_t **Rmsg, uint32_t *Rlen)
{
  uint8_t *msg;
  uint32_t len, slen, tlen;
  tssh_obj_t *R;

  msg= *Rmsg;
  len= *Rlen;
  if (msg+4>end) return NULL;
  
  slen= tssh_get_uint32(msg);
  tlen= slen + 4;
  if (msg+tlen>end) return NULL;
  R= tssh_obj_copy(msg, tlen);
  *Rmsg += tlen;
  *Rlen -= tlen;
  return R;
}


#ifdef TSSH_NO_DEBUG
#define tssh_xparse_uint32(FLD, PB, RET) tssh_parse_uint32(PB,RET)
#define tssh_xparse_uint64(FLD, PB, RET) tssh_parse_uint64(PB,RET)
#define tssh_xparse_string(FLD, PB, RET) tssh_parse_string(PB,RET)
#define tssh_xparse_byte(FLD, PB, RET) tssh_parse_byte(PB, RET)
#define tssh_xparse_bool(FLD, PB, RET) tssh_parse_bool(PB, RET)
#define tssh_xparse_sftpcmd(FLD, PB, RET) tssh_parse_byte(PB, RET)
#else
static int tssh_xparse_uint32(char *fld, tssh_buffer_t* pb, uint32_t *R)
{
  if (tssh_parse_uint32(pb,R)) return 1;
  printf("%s=%d\n",fld,*R);
  return 0;
}
static int tssh_xparse_uint64(char *fld, tssh_buffer_t* pb, uint64_t *R)
{
  if (tssh_parse_uint64(pb,R)) return 1;
  printf("%s=%lld\n",fld,(long long int) *R);
  return 0;
}
static int tssh_xparse_string(char *fld, tssh_buffer_t* pb, tssh_obj_t *R)
{
  if (tssh_parse_string(pb,R)) return 1;
  printf("%s=",fld);
  fwrite(R->data+4, 1, R->len-4, stdout);
  printf("\n");
  return 0;
}

static void tssh_print_cmd_name(uint8_t cmd)
{
  char *p;
  p= NULL;
  switch(cmd)
  {
case SSH_MSG_DISCONNECT: p="disconnect"; break;
case SSH_MSG_IGNORE: p= "ignore"; break;
case SSH_MSG_UNIMPLEMENTED: p= "unimplemented"; break;
case SSH_MSG_DEBUG: p= "debug"; break;
case SSH_MSG_SERVICE_REQUEST: p= "service_request"; break;
case SSH_MSG_SERVICE_ACCEPT: p= "service_accept"; break;
case SSH_MSG_EXT_INFO: p= "ext_info"; break;
case SSH_MSG_KEXINIT: p= "kexinit"; break;
case SSH_MSG_NEWKEYS: p= "newkeys"; break;
case SSH_MSG_KEXDH_INIT: p= "kexdh_init"; break;
case SSH_MSG_KEXDH_REPLY: p= "kexdh_reply"; break;
case SSH_MSG_USERAUTH_REQUEST: p= "userauth_request"; break;
case SSH_MSG_USERAUTH_FAILURE: p= "userauth_failure"; break;
case SSH_MSG_USERAUTH_SUCCESS: p= "userauth_success"; break;
case SSH_MSG_USERAUTH_BANNER: p= "userauth_banner"; break;
case SSH_MSG_USERAUTH_PASSWD_CHANGEREQ: p= "userauth_passwd_changereq"; break;
case SSH_MSG_GLOBAL_REQUEST: p= "global_request"; break;
case SSH_MSG_REQUEST_SUCCESS: p= "request_success"; break;
case SSH_MSG_REQUEST_FAILURE: p= "request_failure"; break;
case SSH_MSG_CHANNEL_OPEN: p= "channel_open"; break;
case SSH_MSG_CHANNEL_OPEN_CONFIRMATION: p= "channel_open_confirmation"; break;
case SSH_MSG_CHANNEL_OPEN_FAILURE: p= "channel_open_failure"; break;
case SSH_MSG_CHANNEL_WINDOW_ADJUST: p= "channel_window_adjust"; break;
case SSH_MSG_CHANNEL_DATA: p= "channel_data"; break;
case SSH_MSG_CHANNEL_EXTENDED_DATA: p= "channel_extended_data"; break;
case SSH_MSG_CHANNEL_EOF: p= "channel_eof"; break;
case SSH_MSG_CHANNEL_CLOSE: p= "channel_close"; break;
case SSH_MSG_CHANNEL_REQUEST: p= "channel_request"; break;
case SSH_MSG_CHANNEL_SUCCESS: p= "channel_success"; break;
case SSH_MSG_CHANNEL_FAILURE: p= "channel_failure"; break;
  }
  if (p) printf("%s\n", p);
   else printf("%d\n", cmd);
}
static int tssh_xparse_byte(char *fld, tssh_buffer_t *pb, uint8_t *R)
{
  if (tssh_parse_byte(pb,R)) return 1;
  printf("%s=", fld);
  if (!strcmp(fld,"cmd")) tssh_print_cmd_name(*R);
                     else printf("%d\n", *R);
  return 0;
}
static int tssh_xparse_bool(char *fld, tssh_buffer_t *pb, uint8_t *R)
{
  if (tssh_parse_bool(pb,R)) return 1;
  printf("%s=%d\n", fld, *R);
  return 0;
}
static void tssh_print_ftp_cmd_name(uint8_t cmd)
{
  char *a;
  switch(cmd)
  {
  case SSH_FXP_INIT: a="init"; break;
  case SSH_FXP_VERSION: a= "version"; break;
  case SSH_FXP_OPEN: a= "open"; break;
  case SSH_FXP_CLOSE: a= "close"; break;
  case SSH_FXP_READ: a= "read"; break;
  case SSH_FXP_WRITE: a= "write"; break;
  case SSH_FXP_LSTAT: a= "lstat"; break;
  case SSH_FXP_FSTAT: a= "fstat"; break;
  case SSH_FXP_SETSTAT: a= "setstat"; break;
  case SSH_FXP_FSETSTAT: a= "fsetstat"; break;
  case SSH_FXP_OPENDIR: a= "opendir"; break;
  case SSH_FXP_READDIR: a= "readdir"; break;
  case SSH_FXP_REMOVE: a= "remove"; break;
  case SSH_FXP_MKDIR: a= "mkdir";
  case SSH_FXP_RMDIR: a= "rmdir";
  case SSH_FXP_REALPATH: a= "realpath"; break;
  case SSH_FXP_STAT: a= "stat"; break;
  case SSH_FXP_RENAME: a= "rename"; break;
  case SSH_FXP_READLINK: a= "readlink"; break;
  case SSH_FXP_LINK: a= "link"; break;
  case SSH_FXP_BLOCK: a= "block"; break;
  case SSH_FXP_UNBLOCK: a= "unblock"; break;
  case SSH_FXP_STATUS: a= "status"; break;
  case SSH_FXP_HANDLE: a= "handle"; break;
  case SSH_FXP_DATA: a= "data"; break;
  case SSH_FXP_NAME: a= "name"; break;
  case SSH_FXP_ATTRS: a= "attrs"; break;
  case SSH_FXP_EXTENDED: a= "extended"; break;
  case SSH_FXP_EXTENDED_REPLY: a= "extended_reply"; break;
  default: printf("%d\n", cmd);return ;
  }
  printf("%s\n",a);
}
static int tssh_xparse_sftpcmd(char *fld, tssh_buffer_t *pb, uint8_t *R)
{
  if (tssh_parse_byte(pb,R)) return 1;
  printf("%s=", fld);
  tssh_print_ftp_cmd_name(*R);
  return 0;
}
#endif




static int tssh_parse_extpair
  (tssh_buffer_t *b, tssh_obj_t *name, tssh_obj_t *data)
{
  if (tssh_parse_string(b, name) || tssh_parse_string(b, data))
    return 1;
  return 0;
}



static int tssh_ftp_parse_file_attr
   (tssh_t *T, tssh_buffer_t *pb, tssh_fileinfo_t *info)
{
  uint32_t flags, count, i, dummy32;
  tssh_obj_t name, value;

  memset(info, 0, sizeof(*info));
  if (tssh_xparse_uint32("fa-flags", pb, &flags)) goto bad_msg;
  if (flags&SSH_FILEXFER_ATTR_SIZE)
  {
    if (tssh_xparse_uint64("fa-size", pb, &info->size)) goto bad_msg;
  }
  if (flags&SSH_FILEXFER_ATTR_UIDGID)
  {
    if (tssh_xparse_uint32("fa-uid", pb, &dummy32)) goto bad_msg;
    if (tssh_xparse_uint32("fa-gid", pb, &dummy32)) goto bad_msg;
  }
  if (flags&SSH_FILEXFER_ATTR_PERMISSIONS)
  {
    if (tssh_xparse_uint32("fa-perms", pb, &info->perm)) goto bad_msg;
  }
  if (flags&SSH_FILEXFER_ATTR_ACMODTIME)
  {
    if (tssh_xparse_uint32("fa-atime", pb, &dummy32)) goto bad_msg;
    if (tssh_xparse_uint32("fa-mtime", pb, &info->mtime)) goto bad_msg;
  }
  if (flags&SSH_FILEXFER_ATTR_EXTENDED)
  {
    if (tssh_xparse_uint32("fa-excount", pb, &count)) goto bad_msg;
    for(i=0;i<count;i++)
    {
      if (tssh_xparse_string("fa-exname", pb, &name)) goto bad_msg;
      if (tssh_xparse_string("fa-exvalue", pb, &value)) goto bad_msg;
    }
  }
  return 0;
bad_msg:
  tssh_log(T, "bad file attributes structure in SFTP response\n");
  return 1;
}

