
static void tssh_put_thread_msg 
    (tssh_queue_t *Q, int reserve, 
     int cmd, void *p1, void *p2, uint32_t i1, void* p3)
{
  tssh_cmd_t *C;
  int p;
  C= malloc(sizeof(*C));
  C->cmd= cmd;
  C->p1= p1;
  C->p2= p2;
  C->i1= i1;
  C->p3= p3;
  pthread_mutex_lock(&Q->mutex);
  while((Q->len+reserve)==Q->size)
    pthread_cond_wait(&Q->have_space, &Q->mutex);
  p= (Q->head+Q->len)%Q->size;
  Q->len++;
  Q->data[p]= C;
  pthread_cond_signal(&Q->have_data, &Q->mutex);
  pthread_mutex_unlock(&Q->mutex);
}

static void tssh_put_thread_msg_simple(tssh_queue_t *Q, int cmd)
{
  tssh_put_thread_msg(Q, 0, cmd, NULL, NULL, 0, NULL);
}

static void tssh_get_thread_msg
  (tssh_queue_t *Q, int *Rcmd, void **Rp1, void **Rp2,
   uint32_t *Ri1, void **Rp3)
{
  tssh_cmd_t *C;
  int p;
  pthread_mutex_lock(&Q->mutex);
  while(Q->len==0)
    pthread_cond_wait(&Q->have_data, &Q->mutex);
  p= Q->head;
  Q->head= (Q->head+1)%Q->size;
  Q->len--;
  C= Q->data[p];
  pthread_cond_broadcast(&Q->have_space, &Q->mutex);
  pthread_mutex_unlock(&Q->mutex);
  *Rcmd= C->cmd;
  *Rp1= C->p1;
  *Rp2= C->p2;
  *Ri1= C->i1;
  *Rp3= C->p3;
  free(C);
}

static void tssh_queue_init(tssh_queue_t *Q)
{
  pthread_mutex_init(&Q->mutex, NULL);
  pthread_cond_init(&Q->have_data, NULL);
  pthread_cond_init(&Q->have_space, NULL);
  Q->len= 0;
  Q->head= 0;
  Q->size= 20;
  Q->data= malloc(sizeof(Q->data[0])*Q->size);
}

static tssh_event_t *tssh_event_new(int type)
{
  tssh_event_t *E;
  E= calloc(1,sizeof(*E));
  E->type= type;
  return E;
}

static void tssh_event_free(tssh_event_t *E)
{
  switch(E->type)
  {
  case TSSH_EVENT_LOG: free(E->D.str); break;
  case TSSH_EVENT_PORT_FORWARD_CONNECT:
    tssh_checkfree(E->connected_address);
    tssh_checkfree(E->originator_address);
    break;
  case TSSH_EVENT_FTP_SUCCESS:
    if (E->fileinfo) tssh_free_finfos(E->fileinfo, E->n_fileinfo);
  case TSSH_EVENT_FTP_FAILURE:
    if (E->ftp_op) tssh_ftop_free(E->ftp_op);
    break;
  }
  free(E);
}


static void tssh_report(tssh_t *T, tssh_event_t *ptr)
{
  int R;
  uint8_t c;
  c= 1;
  do { R= send(T->iface_fds[1], &c,1, 0); } while (R<=0 && errno==EINTR); 
  tssh_put_thread_msg(&T->iface_queue, 0, 1 , ptr, NULL, 0,0);
}

static void tssh_report_string(tssh_t *T, int type, char *str)
{
  tssh_event_t *ev;
  ev= tssh_event_new(type);
  ev->D.str= str;
  tssh_report(T, ev);
}

static void tssh_report_simple(tssh_t *T, int type)
{
  tssh_event_t *ev;
  ev= tssh_event_new(type);
  tssh_report(T, ev);
}

tssh_event_t* tssh_get_event(tssh_t *T)
{
  int R;
  uint8_t c;
  void *ap1, *ap2, *ap3; uint32_t ai1; int cmd;
  do { R= recv(T->iface_fds[0], &c,1, 0); } while (R<=0 && errno==EINTR); 
  tssh_get_thread_msg(&T->iface_queue, &cmd, &ap1, &ap2, &ai1, &ap3);
  if (T->last_event) tssh_event_free(T->last_event);
  T->last_event= NULL;
  return ap1;
}

static void tssh_lock(pthread_mutex_t *mutex)
{
  pthread_mutex_lock(mutex);
}

static void tssh_unlock(pthread_mutex_t *mutex)
{
  pthread_mutex_unlock(mutex);
}

static void tssh_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
{
  pthread_cond_wait(cond, mutex);
}

static void tssh_cond_init(pthread_cond_t *cond)
{
  pthread_cond_init(cond, NULL);
}

static void tssh_mutex_init(pthread_mutex_t *mutex)
{
  pthread_mutex_init(mutex, NULL);
}

static void tssh_cond_signal(pthread_cond_t *cond)
{
  pthread_cond_signal(cond);
}

static void tssh_cond_broadcast(pthread_cond_t *cond)
{
  pthread_cond_broadcast(cond);
}

typedef struct 
{
  void (*func)(void*);
  void *arg;
} tssh_thread_args_t;

static void *tssh_thread_call_func(void *a)
{
  tssh_thread_args_t *args;
  args= a;
  (*args->func)(args->arg);
  free(a);
  return NULL;
}

static void tssh_thread_create(pthread_t *thread, void *func, void *arg)
{
  tssh_thread_args_t *a;
  a= malloc(sizeof(*a));
  a->func= func;
  a->arg= arg;
  pthread_create(thread, NULL, tssh_thread_call_func, a);
}
