
static void tssh_crypto_init_encrypt
   (tssh_crypto_t *t,int type, uint8_t *key, uint8_t *iv)
{
  t->type= type;
  switch(t->type)
  {
  case TSSH_AES256CTR: 
    AES256CTR_init(&t->C.aes256ctr,key,iv);
    break;
  }
}

static void tssh_crypto_init_decrypt
   (tssh_crypto_t *t,int type, uint8_t *key, uint8_t *iv)
{
  t->type= type;
  switch(t->type)
  {
  case TSSH_AES256CTR: 
    AES256CTR_init(&t->C.aes256ctr,key,iv);
    break;
  }
}

static struct tssh_crypto_def tssh_crypto_defs[]=
{
  { "aes256-ctr", TSSH_AES256CTR, 16, 32 },
  { NULL }
};

static struct tssh_crypto_def *tssh_find_crypto_def(char *name)
{
  struct tssh_crypto_def *D;
  for(D=tssh_crypto_defs;D->name;D++)
    if (!strcmp(D->name,name)) return D;
  return NULL;
}


static struct tssh_crypto_def *tssh_nego_crypto_def(tssh_obj_t *serv_o)
{
  tssh_obj_t server_list, server;
  struct tssh_crypto_def *D;

  for(D=tssh_crypto_defs;D->name;D++)
  {
    server_list= *serv_o;
    tssh_namelist_skiplen(&server_list);
    while(tssh_namelist_get(&server_list, &server))
    {
      if (server.len!=strlen(D->name)) continue;
      if (memcmp(server.data,D->name,server.len)) continue;
      return D;
    }
  }
  return NULL;
}                      

static char *tssh_crypto_names()
{
  static char *R;
  int n;
  size_t L;
  buffer_t *B;
  struct tssh_crypto_def *D;

  if (R) return R;
  B= buffer_new(1024);
  for(D=tssh_crypto_defs,n=0;D->name;D++,n++)
  {
    if (n) buffer_write(B, ",", 1);
    buffer_write(B, D->name, strlen(D->name));
  }
  buffer_write(B, "", 1);
  buffer_free(B, (void**) &R, &L);
  return R;
}  


static void tssh_decrypt(tssh_crypto_t *cc, uint8_t *buffer, uint32_t nbytes)
{
  switch(cc->type)
  {
  case TSSH_AES256CTR: 
    AES256CTR_xcrypt(&cc->C.aes256ctr, buffer, nbytes); 
    break;
  }
}

static void tssh_encrypt(tssh_crypto_t *cc, uint8_t *buffer, uint32_t nbytes)
{
  switch(cc->type)
  {
  case TSSH_AES256CTR:
    AES256CTR_xcrypt(&cc->C.aes256ctr, buffer, nbytes); 
    break;
  }
}



