


static void tssh_decode_bytes(tssh_encoder_t *encoder, uint8_t *data, int len)
{
  tssh_decrypt(&encoder->crypto, data, len);
  tssh_hmac_update(&encoder->hmac, data, len);
}


static void tssh_reader_thread(tssh_t *T)
{
  uint8_t  msg_cmd;
  uint32_t packet_len, payload_len, total_len, 
           offset, nblocks, decoded_blocks;
  uint32_t size;
  uint8_t  padding_len;
  tssh_buffer_t *bb;

  int cmd; void *ap1, *ap2, *ap3; uint32_t ai1;

  uint32_t packet_no;

  uint32_t iQ;
  tssh_encoder_t *encoder;
  uint8_t *read_buffer; 
  uint32_t read_buffer_len, read_buffer_size;
  int R;

  packet_no= 0xFFffFFff;
  iQ= 8;
  encoder= NULL;
  read_buffer_size= 1024;
  read_buffer= malloc(read_buffer_size);

message:
  tssh_get_thread_msg(&T->reader_queue, &cmd, &ap1,&ap2, &ai1,&ap3);
  switch(cmd)
  {
  case TSSH_CMD_START_READER:
    memcpy(read_buffer, ap1, ai1);
    read_buffer_len= ai1;
    free(ap1);
    break;
  case TSSH_CMD_QUIT:
    goto done;
  case TSSH_CMD_SETENCODER:
    if (encoder) free(encoder);
    encoder= ap1;
    iQ= encoder->cryptodef->block_len;
    size= encoder->hmac.len;
    if (size<iQ) size= iQ;
    if (size>read_buffer_size)
    {
      read_buffer_size= size;
      read_buffer= realloc(read_buffer, read_buffer_size);
    }
    break;
  }

again:
  packet_no++;
  if (encoder) tssh_hmac_begin(&encoder->hmac, packet_no);
  while (read_buffer_len<iQ) 
  {
    R= tssh_recv(T, read_buffer+read_buffer_len, iQ-read_buffer_len);
    if (R<0) goto fail_recv;
    read_buffer_len+= R;
  }
  if (encoder) tssh_decode_bytes(encoder, read_buffer, iQ);
  read_buffer_len= 0;

  packet_len= tssh_get_uint32(read_buffer);
  padding_len= tssh_get_uint8(read_buffer+4);
  payload_len= packet_len - padding_len - 1;
  total_len= packet_len + 4;

  //printf(" ==== MESSAGE packet_len= %d\n", packet_len);
  //printf(" ==== MESSAGE padding_len= %d\n", padding_len);
  //printf(" ==== MESSAGE payload_len= %d\n", payload_len);

  if ( (total_len%iQ)                  // not a multiple of iQ
       || (padding_len+1>=packet_len)  // empty or negative payload
       || (padding_len<4)              // too short padding
       || (payload_len> (1<<18)) )     // payload too big
     goto fail_length; 

  bb= malloc(sizeof(tssh_buffer_t)+total_len);
  bb->data= (uint8_t*) bb + sizeof(tssh_buffer_t);
  bb->len= total_len;

  memcpy(bb->data, read_buffer, iQ);
  offset= iQ;
  decoded_blocks= 1;

  while(offset<total_len)
  {
    R= tssh_recv(T, bb->data+offset, total_len-offset);
    if (R<0) goto fail_recv;
    offset+= R;
    nblocks= (offset-decoded_blocks*iQ) / iQ;
    if (!nblocks) continue;
    if (encoder)
       tssh_decode_bytes(encoder, bb->data+decoded_blocks*iQ, nblocks*iQ);
    decoded_blocks+= nblocks;
  }

  if (encoder)
  {
    read_buffer_len= 0;
    while(read_buffer_len<encoder->hmac.len)
    {
      R= tssh_recv(T, read_buffer+read_buffer_len,
                           encoder->hmac.len-read_buffer_len);
      if (R<0) goto fail_recv;
      read_buffer_len+= R;
    }
    tssh_hmac_finish(&encoder->hmac);
    if (memcmp(encoder->hmac.data, read_buffer, encoder->hmac.len))
      goto fail_hmac;
    read_buffer_len= 0;
  }

  bb->data+= 5;
  bb->len= payload_len;
  msg_cmd= bb->data[0];
  tssh_put_thread_msg(&T->main_queue,0, TSSH_CMD_NEWPACKET, bb, NULL, 0,0);

  switch(msg_cmd)
  {
  case SSH_MSG_NEWKEYS: 
printf("NEW_KEYS message received by reader, waiting for a new key from main thread\n");
     goto message;
  case SSH_MSG_DISCONNECT: goto done;
  }
  goto again;
done:
  printf("READER QUIT!\n");
  free(read_buffer);
  if (encoder) free(encoder);
  return ;
fail_recv:
  printf("FAIL RECV!\n");
  goto done;
fail_hmac:
  printf("FAIL HMAC!\n");
  goto done;
fail_length:
  printf("FAIL LENGTH!\n");
  goto done;
}

static void* tssh_reader_thread_entry(void *arg)
{
  tssh_reader_thread(arg);
  return NULL;
}


