
static int tssh_kex_compute_K(tssh_t *T)
{
  tssh_mpint_t *F, *K;
  tssh_kex_t *kex;
  uint8_t *k_encoded; uint32_t k_encoded_len;

  kex= T->kex;

  F= tssh_mpint_from_rfc_mpint(kex->F->data);

  if (!tssh_mpint_greater(kex->P, F))
  {
    tssh_mpint_free(F);
    tssh_log(T,"KEXDH_REPLY: given F is greater than prime: aborting\n");
    return 1;
  }
  K= tssh_expmod_by_table(F, kex->X, kex->P_table);
  kex->K= tssh_mpint_to_obj(K);
  tssh_mpint_free(K);
  tssh_mpint_free(F);
  tssh_mpint_free(kex->X);
  kex->X= NULL;
  tssh_numbers_free(kex->P_table);
  kex->P_table= NULL;
#ifdef TSSH_DEBUG_KEX
  tssh_dump_obj("Shared secret", kex->K);
#endif
  return 0;
}


static void tssh_kex_compute_H(tssh_t *T,tssh_hash_t *hh)
{
  tssh_hash_begin(hh);
  tssh_hash_obj(hh, T->o_self_id);
  tssh_hash_obj(hh, T->o_remote_id);
  tssh_hash_obj(hh, T->kex->msg_init_c2s);
  tssh_hash_obj(hh, T->kex->msg_init_s2c);
  tssh_hash_obj(hh, T->kex->host_key);
  tssh_hash_obj(hh, T->kex->E);
  tssh_hash_obj(hh, T->kex->F);
  tssh_hash_obj(hh, T->kex->K);
  tssh_hash_finish(hh);

#ifdef TSSH_DEBUG_KEX
  tssh_debug_hex(hh->data, hh->len, "kex.H");
#endif
  T->kex->H= tssh_obj_copy(hh->data, hh->len);
}

static tssh_obj_t* tssh_genkey
   (tssh_hash_t *hash, tssh_t *T, char* chr, int len)
{
  uint8_t *keydata; int keylen, cp;

  keydata= malloc(len);
  keylen= 0;

  tssh_hash_begin(hash);
  tssh_hash_obj(hash, T->kex->K);
  tssh_hash_obj(hash, T->kex->H);
  tssh_hash_update(hash, chr, 1);
  tssh_hash_obj(hash, T->session_id);
  tssh_hash_finish(hash);

  while(1)
  {
    cp= len-keylen;
    if (cp>hash->len) cp= hash->len;
    memcpy(keydata+ keylen, hash->data, cp);
    keylen+= cp;
    if (keylen==len) break;

    tssh_hash_begin(hash);
    tssh_hash_obj(hash, T->kex->K);
    tssh_hash_obj(hash, T->kex->H);
    tssh_hash_update(hash, keydata, keylen);
    tssh_hash_finish(hash);
  }
#ifdef TSSH_DEBUG_KEX
  tssh_debug_hex(keydata, len, "KEY %s",chr);
#endif
  return tssh_obj_new(keydata,len);
}

