
// this is a curious function. it's used to store some messages
// in string format. these messages are the kex_init payloads
// which are then used in some computations..
static tssh_obj_t *tssh_message_to_str_obj(uint8_t *data,int len)
{
  uint8_t *D;
  D= malloc(len+4);
  memcpy(D+4, data, len);
  tssh_put_uint32(D,len);
  return tssh_obj_new(D,len+4);
}


static void tssh_send_kexinit(tssh_t *T)
{
  int i;
  tssh_opak_t *P;
  char *str;

  T->kex= calloc(1,sizeof(T->kex[0]));

  for(i=0;i<16;i++) T->kex->cookie[i]= rand();

  P= tssh_opak_begin();
  tssh_opak_byte(P, SSH_MSG_KEXINIT);
  tssh_opak_bytes(P, T->kex->cookie, 16);
  tssh_opak_namelist(P, str= tssh_algo_names(tssh_kex_algonames));
  free(str);
  tssh_opak_namelist(P, str= tssh_algo_names(tssh_serverkey_algonames));
  free(str);

  tssh_opak_namelist(P, tssh_crypto_names());
  tssh_opak_namelist(P, tssh_crypto_names());
  tssh_opak_namelist(P, tssh_hmac_names());
  tssh_opak_namelist(P, tssh_hmac_names());

  tssh_opak_namelist(P, str=  tssh_algo_names(tssh_compress_algonames));
  tssh_opak_namelist(P, str);
  free(str);
  tssh_opak_namelist(P, "");
  tssh_opak_namelist(P, "");
  tssh_opak_bool(P, 0);
  tssh_opak_uint32(P, 0);

  uint8_t *Pdata; size_t Plen;
  buffer_collect(P->buffer, (void**) &Pdata, &Plen);
  T->kex->msg_init_c2s= tssh_message_to_str_obj(Pdata,Plen);
  free(Pdata);

  tssh_send_opak(T, P);

  // tssh_dump_obj("kexinit message C=>S", T->kex->msg_init_c2s);
}

static char *oakley_group_14=
      "FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1"
      "29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD"
      "EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245"
      "E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED"
      "EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D"
      "C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F"
      "83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D"
      "670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B"
      "E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9"
      "DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510"
      "15728E5A 8AACAA68 FFFFFFFF FFFFFFFF";

static void tssh_kex_free(tssh_kex_t *kex)
{
}

static int tssh_msg_kexdh_reply(tssh_t *T, tssh_buffer_t *pb)
{
  unsigned char *msg, *end; uint32_t len;
  tssh_opak_t *P;
  tssh_kex_t *kex;
  tssh_encoder_t *encoder_o;
  tssh_hash_t *hash;
  uint8_t dummy;

  if (!T->kex || T->kex->state!=1) return 1;
  T->kex->state= 2;

  kex= T->kex;

  tssh_parse_byte(pb,&dummy);
  kex->host_key= tssh_parse_string_copy(pb);
  kex->F= tssh_parse_mpint_copy(pb);
  kex->signature= tssh_parse_string_copy(pb);

  if (!kex->host_key || !kex->F || !kex->signature)
  {
    tssh_log(T, "KEXDH_REPLY: incomplete message\n");
    return 1;
  }

  if (tssh_kex_compute_K(T)) return 1;
  hash= tssh_kex_hash_alloc(T->kex->kexalgo);
  tssh_kex_compute_H(T,hash);

  // tssh_dump_obj("host key:", T->kex->host_key);
  // tssh_dump_obj("signature:", T->kex->signature);
  if (tssh_verify_signature(T)) 
  {
    tssh_hash_dealloc(hash);
    return 1;
  }

  if (!T->session_id) T->session_id= tssh_obj_copy(kex->H->data, kex->H->len);
  
  struct tssh_crypto_def *cd;
  tssh_obj_t *iv, *key;

  encoder_o= malloc(sizeof(encoder_o[0]));
  encoder_o->cryptodef= kex->cryptodef_c2s;

  cd= kex->cryptodef_c2s;
  iv= tssh_genkey(hash, T, "A", cd->block_len); 
  key= tssh_genkey(hash, T, "C", cd->key_len);
  tssh_crypto_init_encrypt(&encoder_o->crypto, cd->type,
                           key->data, iv->data);
  tssh_obj_free_deep(key);
  tssh_obj_free_deep(iv);

  key= tssh_genkey(hash, T, "E", kex->macdef_c2s->key_len);
  tssh_hmac_init(&encoder_o->hmac, kex->macdef_c2s->type, key->data);
  tssh_obj_free_deep(key);

  kex->encoder_i= malloc(sizeof(kex->encoder_i[0]));
  kex->encoder_i->cryptodef= kex->cryptodef_s2c;

  cd= kex->cryptodef_s2c;
  iv= tssh_genkey(hash,T, "B", cd->block_len);
  key= tssh_genkey(hash, T, "D", cd->key_len);
  tssh_crypto_init_decrypt(&kex->encoder_i->crypto, cd->type,
                           key->data, iv->data);
  tssh_obj_free_deep(key);
  tssh_obj_free_deep(iv);

  key= tssh_genkey(hash, T, "F", kex->macdef_s2c->key_len);
  tssh_hmac_init(&kex->encoder_i->hmac, kex->macdef_s2c->type, key->data);
  tssh_obj_free_deep(key);

  tssh_hash_dealloc(hash);

  P= tssh_opak_begin();
  tssh_opak_byte(P, SSH_MSG_NEWKEYS );
  tssh_send_opak(T, P);
  
  T->oQ= encoder_o->cryptodef->block_len;
  tssh_put_thread_msg(&T->writer_queue, 0, TSSH_CMD_SETENCODER, encoder_o,
                      NULL, 0,0);
  return 0;
}

static void tssh_send_kexdh_init(tssh_t *T)
{
  int i,j;
  tssh_opak_t *O;
  tssh_mpint_t *P,*X, *two, *E;
  tssh_kex_t *kex;

  P= tssh_mpint_from_hexstr(oakley_group_14);
  X= tssh_mpint_zero(P->len);
  kex= T->kex;

  for(i=0;i<P->len;i++) 
  {
    for(j=0;j<TSSH_WORD_SIZE/8;j++)
       X->data[i] |= (rand()%0xff) << (j*8);
  }
  // this makes it smaller than the prime
  if ((X->data[X->len-1]+1) == 0) X->data[X->len-1] ^= 1;
  kex->X= X;
  kex->P= P;
  kex->P_table= tssh_multiples_of_modulus(P->data, P->len);

  two= tssh_mpint_from_uint32(2);
  E= tssh_expmod_by_table(two, kex->X, kex->P_table);
  tssh_mpint_free(two);

#ifdef TSSH_DEBUG_KEX
  printf("-------DH INIT---------\n");
  tssh_mpint_print("x= ", X);
  tssh_mpint_print("p= ", P);
  tssh_mpint_print("e= ", E);
  printf("-------END---------\n");
#endif

  kex->E= tssh_mpint_to_obj(E);
  tssh_mpint_free(E);

  O= tssh_opak_begin();
  tssh_opak_byte(O, SSH_MSG_KEXDH_INIT);
  tssh_opak_obj(O, kex->E);
  tssh_send_opak(T, O);
}

// process incoming kexinit message
static int tssh_msg_kexinit(tssh_t *T, tssh_buffer_t *pb)
{
  uint8_t *msg,*end; uint32_t len;
  uint8_t *cookie, cmd;
  tssh_obj_t dummy_o;
  uint8_t have_first_kex;
  uint32_t dummy_i;
  tssh_obj_t names;
  void *p; int k;
  char *str;
  char *fn;

  if (!T->kex) tssh_send_kexinit(T);
  if (T->kex->state!=0) return 1;
  T->kex->state= 1;

  T->kex->msg_init_s2c= tssh_message_to_str_obj(pb->data, pb->len);

  fn= "Command";
  if (tssh_parse_byte(pb, &cmd)) goto f_parse;
  fn= "Cookie";
  if (tssh_parse_bytes(pb, &cookie, 16)) goto f_parse;

  fn= "KEX Algorithm";
  if (tssh_parse_namelist(pb, &names)) goto f_parse;
  k= T->kex->kexalgo= tssh_nego_algo(&names, tssh_kex_algonames);
  if (k<0) goto f_nego;

  fn= "Server Host Key Algorithm";
  if (tssh_parse_namelist(pb, &names)) goto f_parse;
  k= T->kex->serverkeyalgo= tssh_nego_algo(&names, tssh_serverkey_algonames);
  if (k<0) goto f_nego;

  fn= "Encryption C->S"; 
  if (tssh_parse_namelist(pb,  &names)) goto f_parse;
  p= T->kex->cryptodef_c2s= tssh_nego_crypto_def(&names);
  if (!p) goto f_nego;

  fn= "Encryption S->C"; 
  if (tssh_parse_namelist(pb,  &names)) goto f_parse;
  p= T->kex->cryptodef_s2c= tssh_nego_crypto_def(&names);
  if (!p) goto f_nego;

  fn= "MAC C->S";
  if (tssh_parse_namelist(pb,  &names)) goto f_parse;
  p= T->kex->macdef_c2s= tssh_nego_hmac_def(&names);
  if (!p) goto f_nego;

  fn= "MAC S->C";
  if (tssh_parse_namelist(pb,  &names)) goto f_parse;
  p= T->kex->macdef_s2c= tssh_nego_hmac_def(&names);
  if (!p) goto f_nego;
  
  fn= "Compression C->S";
  if (tssh_parse_namelist(pb,  &names)) goto f_parse;
  k= T->kex->compalgo_c2s= tssh_nego_algo(&names, tssh_compress_algonames);
  if (k<0) goto f_nego;

  fn= "Compression S->C";
  if (tssh_parse_namelist(pb,  &names)) goto f_parse;
  k= T->kex->compalgo_s2c= tssh_nego_algo(&names, tssh_compress_algonames);
  if (k<0) goto f_nego;

  //
  // below fields aren't actually used, but I will parse them just
  // just for checking the validity of the message
  //
  fn= "Lang C->S";
  if (tssh_parse_namelist(pb, &dummy_o)) goto f_parse;
  fn= "Lang S->C";
  if (tssh_parse_namelist(pb, &dummy_o)) goto f_parse;
  fn= "first_kex_follows";
  if (tssh_parse_bool(pb, &have_first_kex)) goto f_parse;
  fn= "reserved";
  if (tssh_parse_uint32(pb, &dummy_i)) goto f_parse; 

  tssh_send_kexdh_init(T);
  return 0;
f_parse:
  tssh_log(T, "kexinit(): error parsing field %s\n", fn);
  return 1;
f_nego:
  tssh_log(T, "unable to negotiate field %s\n"
              "Server Value: %S\n", fn, &names);
  return 1;
}


static int tssh_msg_newkeys(tssh_t *T, tssh_buffer_t *pb)
{
  tssh_opak_t *O;
  if (!T->kex || T->kex->state!=2) return 1;
  tssh_put_thread_msg(&T->reader_queue, 0,
             TSSH_CMD_SETENCODER, T->kex->encoder_i, NULL,0,0);
  T->kex->encoder_i= NULL;
  tssh_kex_free(T->kex);
  T->kex= NULL;
  T->have_keys= 1;
  tssh_send_ignore(T);
  if (T->auth) return 0;
  O= tssh_opak_begin();
  tssh_opak_byte(O, SSH_MSG_SERVICE_REQUEST);
  tssh_opak_string(O, "ssh-userauth");
  tssh_send_opak(T,O);
  T->auth_state= 1;
  return 0;
}

