

static void tssh_channel_delete(tssh_t *T, tssh_channel_t *ch)
{
  int i;
  for(i=0;i<T->nchannels;i++)
    if (T->channels[i]==ch)
    {
      T->channels[i]= T->channels[T->nchannels-1];
      T->nchannels--;
      break;
    }
  if (ch->env)
  {
    for(i=0;ch->env[i];i++) free(ch->env[i]);
    free(ch->env);
  }
  memset(ch, 0, sizeof(*ch));
  free(ch);
}
static tssh_opak_t *tssh_channel_open_opak(char *type, tssh_channel_t *ch)
{
  tssh_opak_t *O;
  O= tssh_opak_begin();
  tssh_opak_byte(O, SSH_MSG_CHANNEL_OPEN);
  tssh_opak_string(O, type);
  tssh_opak_uint32(O, ch->local_number);
  tssh_opak_uint32(O, ch->window_size_s2c);
  tssh_opak_uint32(O, ch->max_packet_size);
  return O;
}

static tssh_channel_t *tssh_find_channel(tssh_t *T, uint32_t local_id)
{
  int i;
  for(i=0;i<T->nchannels;i++)
   if (T->channels[i]->local_number==local_id)
     return T->channels[i];
  tssh_log(T, "can't find channel #%d\n", local_id);
  return NULL;
}

static tssh_opak_t *tssh_channel_request
  (tssh_channel_t *ch, int want_reply, char *req)
{
  tssh_opak_t *O;
  O= tssh_opak_begin();
  tssh_opak_byte(O, SSH_MSG_CHANNEL_REQUEST);
  tssh_opak_uint32(O, ch->remote_number);
  tssh_opak_string(O, req);
  tssh_opak_bool(O, want_reply ? 1: 0);
  return O;
}


static tssh_channel_t *tssh_channel_create(tssh_t *T,int type)
{
  tssh_channel_t *ch;
  if (T->avail_channel_no==0) return NULL;
  ch= calloc(1,sizeof(*ch));
  ch->type= type;
  ch->local_number= T->avail_channel_no++;
  if (T->nchannels==T->schannels)
  {
    T->schannels+= T->schannels/2 + 10;
    T->channels= realloc(T->channels, sizeof(T->channels[0])*T->schannels);
  }
  T->channels[T->nchannels++]= ch;
  ch->window_size_s2c= TSSH_INIT_WINDOW_SIZE;
  ch->max_packet_size= TSSH_MAX_CHANNEL_PACKET_SIZE;
  ch->transport= T;
  ch->pend_max= TSSH_CHANNEL_PENDING_MAX;

  tssh_mutex_init(&ch->pend_mutex);
  tssh_cond_init(&ch->pend_cv);
  return ch;
}

static void tssh_send_window_adjust
  (tssh_t *T, tssh_channel_t *ch, uint32_t len)
{
  tssh_opak_t *O;
  O= tssh_opak_begin();
  tssh_opak_byte(O, SSH_MSG_CHANNEL_WINDOW_ADJUST);
  tssh_opak_uint32(O, ch->remote_number);
  tssh_opak_uint32(O, len);
  tssh_send_opak(T, O);
}

// this gets called after the header part of
//    SSH_MSG_CHANNEL_DATA or
//    SSH_MSG_CHANNEL_EXTENDED_DATA 
// is parsed.
static int tssh_process_channel_data
    (tssh_t *T, uint32_t localid, int event_code, tssh_buffer_t **Rpb)
{
  uint8_t cmd;
  tssh_obj_t dataobj;
  tssh_channel_t *ch;
  tssh_buffer_t *pb;
  tssh_event_t *E;

  pb= *Rpb;
  if (tssh_parse_string(pb, &dataobj))
  {
    tssh_log(T, "bad channel data message\n");
    return 1;
  }

  ch= tssh_find_channel(T, localid);
  // let data for an unknown channel be simply ignored, maybe we lost
  // something along the way.
  if (!ch) return 0;

  pb->data= dataobj.data+4;
  pb->len= dataobj.len-4;

  if (ch->window_size_s2c<pb->len)
  {
    tssh_log(T, "channel %d doesn't have enough window space for incoming "
                "message with length= %d. window size is= %d\n",
                  localid, pb->len, ch->window_size_s2c);
    // this is a potentially dangerous error, let the connection drop
    return 1;
  }

  ch->window_size_s2c -= pb->len;
  if (ch->window_size_s2c<ch->max_packet_size)
  {
    tssh_send_window_adjust(T, ch, TSSH_INIT_WINDOW_SIZE-ch->window_size_s2c);
    ch->window_size_s2c= TSSH_INIT_WINDOW_SIZE;
  }

  *Rpb= NULL; // this prevents the pb from being freed
  if (ch==T->ftp_channel)
    return tssh_ftp_channel_data(T, pb);

  E= tssh_event_new(TSSH_EVENT_CHANNEL_DATA);
  E->buffer= pb;
  E->channel= ch;
  tssh_report(T,E);

  return 0;
}

static int tssh_msg_channel_data(tssh_t *T, tssh_buffer_t **Rpb)
{
  uint8_t cmd;
  uint32_t localid;
  tssh_buffer_t *pb;

  pb= *Rpb;
  if (tssh_xparse_byte("cmd", pb, &cmd) ||
      tssh_xparse_uint32("localid", pb, &localid))
  {
    tssh_log(T, "bad channel data message\n");
    return 1;
  }

  return tssh_process_channel_data(T, localid, TSSH_EVENT_CHANNEL_DATA, Rpb);
}

static int tssh_msg_channel_extended_data(tssh_t *T, tssh_buffer_t **Rpb)
{
  uint8_t cmd;
  uint32_t localid;
  uint32_t data_type_code;
  tssh_buffer_t *pb;

  pb= *Rpb;
  if (tssh_xparse_byte("cmd", pb, &cmd) ||
      tssh_xparse_uint32("localid", pb, &localid) ||
      tssh_xparse_uint32("datacode", pb, &data_type_code))
  {
    tssh_log(T, "bad channel data message\n");
    return 1;
  }
  if (data_type_code!=SSH_EXTENDED_DATA_STDERR)
  {
    tssh_log(T, "unknown data_type_code %d in channel_extended_data for %d\n",
                  data_type_code, localid);
    return 0;  // ignore the stuff
  }
  return tssh_process_channel_data(T, localid, 
                                      TSSH_EVENT_CHANNEL_EXTENDED_DATA, Rpb);
}


static int tssh_msg_window_adjust(tssh_t *T, tssh_buffer_t *pb)
{
  uint8_t cmd;
  uint32_t localid, amount;
  tssh_channel_t *ch;

  if (tssh_xparse_byte("cmd", pb, &cmd) ||
      tssh_xparse_uint32("localid", pb, &localid) ||
      tssh_xparse_uint32("amount", pb, &amount))
  {
    tssh_log(T, "bad window_adjust message\n");
    return 1;
  } 

  ch= tssh_find_channel(T, localid);
  if (!ch) return 0;

  ch->window_size_c2s += amount;
  return ch==T->ftp_channel ? tssh_forward_ftp_write(T) : 
                              tssh_forward_channel_write(T, ch);
}

static tssh_write_buffer_t *tssh_alloc_write_buffer(uint32_t size)
{
  tssh_write_buffer_t *WB;
  WB= malloc(sizeof(*WB) + size + TSSH_CHANNEL_DATA_HDRSIZE);
  WB->size= size;
  WB->len= 0;
  //
  // WB->data always points to a position where there is enough
  // space before for the header to go..
  //
  WB->data= (uint8_t*) WB + sizeof(*WB) + TSSH_CHANNEL_DATA_HDRSIZE;
  WB->next= NULL;
  return WB;
}

int tssh_channel_write_buffer(tssh_channel_t *ch, tssh_write_buffer_t *B)
{
  tssh_t *T;
  T= ch->transport;
    // B->len will be used to track the amount of remaining data
    // after each send. B->size will be subtracted from the amount
    // of buffered data when the buffer is deallocated.
  B->size= B->len;  
  tssh_lock(&ch->pend_mutex);
  while( (B->len+ch->pend_size > ch->pend_max) && (!ch->close_sent)) 
    tssh_cond_wait(&ch->pend_cv, &ch->pend_mutex);
  if (ch->close_sent)
  {
    tssh_unlock(&ch->pend_mutex);
    return 1;
  }
  if (!ch->pend_last) ch->pend_first= B;
            else ch->pend_last->next= B;
  ch->pend_last= B;
  ch->pend_size += B->len;
  tssh_unlock(&ch->pend_mutex);
  tssh_put_thread_msg(&T->main_queue, 0, TSSH_CMD_CHANNEL_WRITE, ch, NULL, 0,0);
  return 0;
}

int tssh_channel_write(tssh_channel_t *ch, uint8_t *data, uint32_t len)
{
  tssh_write_buffer_t *B;
  B= tssh_alloc_write_buffer(len);
  B->len= len;
  memcpy(B->data, data, len);
  if (tssh_channel_write_buffer(ch, B))
  {
    free(B);
    return 1;
  }
  return 0;
}

static uint32_t tssh_outbound_channel_data_packet_size
    (uint32_t oQ, uint32_t data_len, uint8_t *Rpadding)
{
   uint8_t padding_len;
   uint32_t total_size;
   padding_len= 4;
   total_size= TSSH_CHANNEL_DATA_HDRSIZE + data_len + padding_len;
   if (total_size%oQ)
      padding_len += oQ - total_size%oQ;
   if (Rpadding) *Rpadding= padding_len;
   return TSSH_CHANNEL_DATA_HDRSIZE + data_len + padding_len;
}

static int tssh_forward_channel_write(tssh_t *T, tssh_channel_t *ch)
{
  tssh_event_t *E;
  tssh_write_buffer_t *B, *buffer_to_free;
  uint32_t size, packet_size;
  uint32_t diff;
  uint8_t *odata;

  if (ch->window_size_c2s==0) return 0;

  tssh_lock(&ch->pend_mutex);
  B= ch->pend_first;

  if (!B) 
  {
    tssh_unlock(&ch->pend_mutex);
    return 0;
  }

  size= tssh_min_uint32(B->len, ch->window_size_c2s);
  packet_size= tssh_outbound_channel_data_packet_size(T->oQ, size, NULL);
  if (packet_size > ch->max_packet_size)
  {
    diff= packet_size - ch->max_packet_size;
    if (diff >= size)
    {
      // this means that a packet of max_packet_size is unable to hold even
      // one byte of the channel data payload. Let's say
      // max_packet_size = 16.
      // we already need 14 bytes to hold the header, and at least 4 bytes
      // for padding. so, even the header doesn't fit into the biggest
      // possible packet.
      // For 5 bytes of channel data, if we ended up with 
      // packet_size = 32 and max_packet_size= 16,
      // we could reduce the amount of data by the difference,
      // but this doesn't work since we only have 5 bytes of data.
      // So, the above test correctly identifies the situation.
       tssh_unlock(&ch->pend_mutex);
       tssh_log(T, 
          "maximum packet size for channel %d is too small: %d bytes\n",
          ch->local_number, ch->max_packet_size);
       // there is something seriously wrong with this server
       return 1; 
    }
    size -= diff;
  }
  ch->window_size_c2s -= size;

  odata= B->data;
  B->data+= size;
  B->len-= size;
  E= NULL;
  buffer_to_free= NULL;
  if (B->len==0)
  {
    ch->pend_first= B->next;
    if (!ch->pend_first) ch->pend_last= NULL;
    ch->pend_size -= B->size;
    buffer_to_free= B;
    tssh_cond_signal(&ch->pend_cv);
    E= tssh_event_new(TSSH_EVENT_CHANNEL_WRITE_OK);
    E->channel= ch;
  }
  tssh_unlock(&ch->pend_mutex);

  tssh_put_thread_msg(&T->writer_queue, 0, TSSH_CMD_CHANNEL_WRITE, ch,
                             odata, size, buffer_to_free);

  if (E) tssh_report(T, E);
  return 0;
}

static int tssh_channel_write_ok
   (tssh_t *T, tssh_channel_t *ch, uint32_t size)
{
  if (!ch) return 0;
  if (ch==T->ftp_channel) 
  { 
    if (tssh_forward_ftp_write(T)) return 1;
  }
  else
  {
    if (tssh_forward_channel_write(T, ch)) return 1; 
  }
  return 0;
}


static void tssh_free_channel_write_buffers(tssh_channel_t *ch)
{
  tssh_write_buffer_t *B,*next;
  for(B=ch->pend_first;B;B=next)
  {
    next= B->next;
    free(B);
  }
  ch->pend_first= ch->pend_last= NULL;
}

static void tssh_remove_channel(tssh_t *T, tssh_channel_t *ch)
{
  int i;
  for(i=0;i<T->nchannels;i++)
    if (T->channels[i]==ch)
    {
      T->channels[i]= T->channels[T->nchannels-1];
      T->nchannels--;
      return ;
    }
}

static void tssh_send_channel_close(tssh_t *T, tssh_channel_t *ch)
{
  tssh_opak_t *O;

  if (ch->close_sent) return ;

  O= tssh_opak_begin();
  tssh_opak_byte(O, SSH_MSG_CHANNEL_CLOSE);
  tssh_opak_uint32(O, ch->remote_number);
  tssh_send_opak(T, O);
  ch->close_sent= 1;

  tssh_remove_channel(T, ch);

  // now rip up the pending data.
  tssh_lock(&ch->pend_mutex);
  tssh_free_channel_write_buffers(ch);

  // we signal any pending write requests so that they return
  // with an error code. 
  tssh_cond_signal(&ch->pend_cv);
  tssh_unlock(&ch->pend_mutex);

  // we now report the event to the user so that they can
  // call destroy() to get rid of the memory.
  if (ch!=T->ftp_channel)
    tssh_report_channel(T, TSSH_EVENT_CHANNEL_CLOSED, ch);
}

static int tssh_msg_channel_close(tssh_t *T, tssh_buffer_t *pb)
{
  uint8_t cmd; uint32_t localid;
  tssh_channel_t *ch;

  if (tssh_xparse_byte("cmd",pb,&cmd) ||
      tssh_xparse_uint32("localid", pb, &localid))
  {
    tssh_log(T, "bad channel_close message\n");
    return 1;
  }
  ch= tssh_find_channel(T, localid);
  if (!ch) return 0;
  if (!ch->close_sent)
    tssh_send_channel_close(T, ch);
  if (ch==T->ftp_channel)
    tssh_ftp_channel_closed(T);
  return 0;
}


void tssh_channel_destroy(tssh_channel_t *ch)
{
  free(ch);
}

