

static void tssh_shell_start(tssh_t *T, char **env)
{
  tssh_channel_t *ch;
  tssh_opak_t *O;
  ch= tssh_channel_create(T, TSSH_CHANNEL_SHELL);
  ch->env= env;
  O= tssh_channel_open_opak("session", ch);
  tssh_send_opak(T, O);
}

static int tssh_shell_channel_success(tssh_t *T, tssh_channel_t *ch)
{
  tssh_opak_t *O;
  tssh_event_t *E;
  // state 0: channel open was sent
  // state 1: waiting for the response to a pty request or 
  //          environment variable request 
  // state 2: shell request was sent
  // state 100: all ok

  switch(ch->state)
  {
  case 0: // this is impossible, we got a success message before
          // we ever sent a request. anyway, let's ignore it for
          // now and pretend that this never happens.
  case 1: goto continue_environ;
  case 2: goto shell_ok;
  }
  tssh_log(T, "internal error: inconsistent channel state\n");
  return 1;
continue_environ:
  if (!ch->env || !ch->env[ch->sent_env] || !ch->env[ch->sent_env+1])
     goto send_shell_req;
  O= tssh_channel_request(ch, 1, "env");
  tssh_opak_string(O, ch->env[ch->sent_env]);
  tssh_opak_string(O, ch->env[ch->sent_env+1]);
  tssh_send_opak(T, O);
  ch->sent_env+= 2;
  ch->state= 1;
  return 0;
send_shell_req:
  O= tssh_channel_request(ch, 1, "shell");
  tssh_send_opak(T, O);
  ch->state= 2;
  return 0;
shell_ok:
  E= tssh_event_new(TSSH_EVENT_SHELL_OK);
  E->channel= ch;
  tssh_report(T, E);
  ch->state= 100;
  return 0;
}

static int tssh_shell_channel_failure(tssh_t *T, tssh_channel_t *ch)
{
  tssh_event_t *E;
  E= tssh_event_new(TSSH_EVENT_SHELL_FAILURE);
  tssh_report(T, E);
  tssh_channel_delete(T, ch);
  return 0;
}

void tssh_shell(tssh_t *T, char **env)
{
  char **ecopy;
  int i, n;
  if (!env) { ecopy= NULL; goto done; }
  for(n=0;env[n];n++) ;
  ecopy= malloc(sizeof(ecopy[0])*(n+1));
  for(i=0;i<n;i++) ecopy[i]= strdup(env[i]);
  ecopy[i]= NULL;
done:
  tssh_put_thread_msg(&T->main_queue, 0, TSSH_CMD_SHELL, ecopy, NULL, 0,0);
}

static void tssh_shell_request_pty(tssh_t *T, tssh_channel_t *ch)
{
  tssh_opak_t *O;
  static const uint8_t terminal_modes[]=
  {
    0, 0, 0, 1,
    0,
  };
  O= tssh_channel_request(ch,1, "pty-req");
  tssh_opak_string(O, "vt100");
  tssh_opak_uint32(O, 80);
  tssh_opak_uint32(O, 24);
  tssh_opak_uint32(O, 640);
  tssh_opak_uint32(O, 480);
  tssh_opak_bytes(O, (uint8_t*) terminal_modes, sizeof(terminal_modes));
  tssh_send_opak(T, O);
  ch->state= 1;
}
