
static void tssh_send_port_forward_request(tssh_t *T, uint32_t port)
{
  int i;
  tssh_opak_t *O;
  if (port)
  {
    for(i=0;i<T->nfwreq;i++) 
      if (T->fwreq[i]==port)  // already in use
         return ;
  }
  O= tssh_opak_begin();
  tssh_opak_byte(O, SSH_MSG_GLOBAL_REQUEST);
  tssh_opak_string(O, "tcpip-forward");
  tssh_opak_bool(O, 1);
  tssh_opak_string(O, "localhost");
  tssh_opak_uint32(O, port);
  tssh_send_opak(T, O);
  tssh_record_global_request(T, TSSH_GR_PORT_FORWARD, NULL, port);
  tssh_array_grow(&T->a_fwreq);
  T->fwreq[T->nfwreq++]= port;
}

void tssh_request_port_forward(tssh_t *T, uint32_t port)
{
  tssh_put_thread_msg(&T->main_queue, 0, TSSH_CMD_PORT_FORWARD,
                                         NULL, NULL, port, 0);
}

static int tssh_port_forward_success
   (tssh_t *T, uint32_t reqport, tssh_buffer_t *pb)
{
  tssh_event_t *E;
  uint8_t cmd; uint32_t port;
  tssh_parse_byte(pb, &cmd);
  if (!reqport)
  {
    int i;
    if (tssh_xparse_uint32("port", pb, &port))
    {
      tssh_log(T, "bad MSG_SUCCESS for port forward request\n");
      return 1;
    }
    for(i=0;i<T->nfwreq;i++)
      if (T->fwreq[i]==0) { T->fwreq[i]= port; break; }
    reqport= port;
  }
  E= tssh_event_new(TSSH_EVENT_PORT_FORWARD_SUCCESS);
  E->port= reqport;
  tssh_report(T, E);
  return 0;
}

static int tssh_port_forward_failure(tssh_t *T, uint32_t reqport)
{
  int i,j;
  tssh_event_t *E;
  for(i=0;i<T->nfwreq;i++)
    if (T->fwreq[i]==reqport) 
    {
      for(j=i;j+1<T->nfwreq;j++) T->fwreq[j]= T->fwreq[j+1];
      T->nfwreq--;
      break;
    }
  E= tssh_event_new(TSSH_EVENT_PORT_FORWARD_FAILURE);
  E->port= reqport;
  tssh_report(T, E);
  return 0;
}

static void tssh_send_channel_open_failure
    (tssh_t *T, uint32_t remote_id, uint32_t reason, char *msg)
{
  tssh_opak_t *O;
  O= tssh_opak_begin();
  tssh_opak_byte(O, SSH_MSG_CHANNEL_OPEN_FAILURE);
  tssh_opak_uint32(O, remote_id);
  tssh_opak_uint32(O, reason);
  tssh_opak_string(O, msg);
  tssh_opak_string(O, "en");
  tssh_send_opak(T, O);
}

static int tssh_init_port_forward_channel
  (tssh_t *T, tssh_buffer_t *pb, tssh_channel_t *ch)
{
  tssh_event_t *E;
  tssh_obj_t coaddr, oraddr;
  uint32_t coport, orport;
  int i;
  if (tssh_xparse_string("coaddr",pb, &coaddr) ||
      tssh_xparse_uint32("coport",pb, &coport) ||
      tssh_xparse_string("oraddr",pb, &oraddr) ||
      tssh_xparse_uint32("orport",pb, &orport))
  {
    tssh_log(T, "bad channel_open for port forwarding\n");
    return 1; 
  }
  for(i=0;i<T->nfwreq;i++)
    if (T->fwreq[i]==coport) break;
  if (i==T->nfwreq) 
  {
    tssh_log(T,"channel_open: port %d was not requested\n", coport);
    return 1;
  }
  E= tssh_event_new(TSSH_EVENT_PORT_FORWARD_CONNECT);
  E->channel= ch;
  E->connected_port= coport;
  E->connected_address= tssh_xstr_to_str(&coaddr);
  E->originator_port= orport;
  E->originator_address= tssh_xstr_to_str(&oraddr);
  tssh_report(T,E);
  return 0;
}

 // this happens only in the case of forwarded connections
 // we open the channels in the other way around for everything else.
static int tssh_msg_channel_open(tssh_t *T, tssh_buffer_t *pb)
{
  tssh_channel_t *ch;
  tssh_opak_t *O;
  uint8_t cmd;
  uint32_t remote_id, init_win, max_packet_size;
  tssh_obj_t type_str;
  uint32_t type;

  if (tssh_parse_byte(pb, &cmd) ||
      tssh_xparse_string("type", pb, &type_str) ||
      tssh_xparse_uint32("remoteid", pb, &remote_id) ||
      tssh_xparse_uint32("initwin", pb, &init_win) ||
      tssh_xparse_uint32("maxpack", pb, &max_packet_size))
  {
    tssh_log(T, "bad channel_open msg\n");
    return 1;
  }  

  if (!tssh_xstr_compare(&type_str, "forwarded-tcpip"))
    type= TSSH_CHANNEL_PORT_FORWARD;
  else goto unknown_channel_type;

  ch= tssh_channel_create(T, type);
  ch->remote_number= remote_id;
  ch->max_packet_size= tssh_min_uint32(ch->max_packet_size, max_packet_size);
  ch->window_size_c2s= init_win;

  switch(type)
  {
  case TSSH_CHANNEL_PORT_FORWARD: 
    if (tssh_init_port_forward_channel(T, pb, ch)) goto failed_init;
    break;
  }


  O= tssh_opak_begin();
  tssh_opak_byte(O, SSH_MSG_CHANNEL_OPEN_CONFIRMATION);
  tssh_opak_uint32(O, remote_id);
  tssh_opak_uint32(O, ch->local_number);
  tssh_opak_uint32(O, ch->window_size_s2c);
  tssh_opak_uint32(O, ch->max_packet_size);
  tssh_send_opak(T, O);
  return 0;

failed_init:
  // for some reason, the channel was not initialized properly.
  // let's just send a failure message and be cool with it, even
  // though the original failure could be caused by a bad message.
  tssh_remove_channel(T, ch);
  free(ch);
  tssh_log(T,"incoming channel_open: unable to initialize (%S)\n",&type_str);
  tssh_send_channel_open_failure(T, remote_id,
                    SSH_OPEN_RESOURCE_SHORTAGE, "init failed");
  return 0;

unknown_channel_type:
  tssh_log(T, "incoming channel_open: unknown channel type %S\n", &type_str);
  tssh_send_channel_open_failure(T, remote_id, 
                    SSH_OPEN_UNKNOWN_CHANNEL_TYPE, "unknown channel type");
  return 0;
}

static void tssh_port_forward_eof(tssh_t *T, tssh_channel_t *ch)
{
  tssh_send_channel_close(T,ch);
}

