
//
// this is a copy of the forward_channel_write, but without the
// reports to the user or the locks.
//
static int tssh_forward_ftp_write(tssh_t *T)
{
  tssh_event_t *E;
  tssh_write_buffer_t *B, *buffer_to_free;
  uint32_t size, packet_size;
  uint32_t diff;
  uint8_t *odata;
  tssh_channel_t *ch;

  ch= T->ftp_channel;
  if (!ch) return 0;
  if (ch->window_size_c2s==0) return 0;

  B= ch->pend_first;
  if (!B) return 0;

  size= tssh_min_uint32(B->len, ch->window_size_c2s);
  packet_size= tssh_outbound_channel_data_packet_size(T->oQ, size, NULL);
  if (packet_size > ch->max_packet_size)
  {
    diff= packet_size - ch->max_packet_size;
    if (diff >= size)
    {
      // this means that a packet of max_packet_size is unable to hold even
      // one byte of the channel data payload. Let's say
      // max_packet_size = 16.
      // we already need 14 bytes to hold the header, and at least 4 bytes
      // for padding. so, even the header doesn't fit into the biggest
      // possible packet.
      // For 5 bytes of channel data, if we ended up with 
      // packet_size = 32 and max_packet_size= 16,
      // we could reduce the amount of data by the difference,
      // but this doesn't work since we only have 5 bytes of data.
      // So, the above test correctly identifies the situation.
       tssh_log(T, 
          "maximum packet size for channel %d is too small: %d bytes\n",
          ch->local_number, ch->max_packet_size);
       // there is something seriously wrong with this server
       return 1; 
    }
    size -= diff;
  }
  ch->window_size_c2s -= size;

  odata= B->data;
  B->data+= size;
  B->len-= size;
  E= NULL;
  buffer_to_free= NULL;
  if (B->len==0)
  {
    ch->pend_first= B->next;
    if (!ch->pend_first) ch->pend_last= NULL;
    ch->pend_size -= B->size;
    buffer_to_free= B;
  }

  tssh_put_thread_msg(&T->writer_queue, 0, TSSH_CMD_CHANNEL_WRITE, ch,
                             odata, size, buffer_to_free);
  return 0;
}

static tssh_write_buffer_t* tssh_ftp_cmd_formatp
   (tssh_t *T,const char *fmt, va_list aq)
{
  int i;
  va_list ap;
  uint32_t size;
  tssh_write_buffer_t *wb;

  uint32_t v_int; uint64_t v_qua; uint16_t v_wor;
  char *v_str;
  uint8_t *v_bytp;
  uint8_t v_byt;
  tssh_obj_t *v_obj;

  wb= NULL;

again:
  size= 4;
  va_copy(ap,aq);
  if (wb) tssh_wb_put_uint32(wb, 0);
  for(i=0;fmt[i];i++)
  switch(fmt[i])
  {
  case 's': 
    v_str= va_arg(ap, char*);
    if (wb) tssh_wb_put_string(wb, v_str);
    size+= strlen(v_str) + 4;
    break;
  case 'd':
    v_int= va_arg(ap, uint32_t);
    if (wb) tssh_wb_put_uint32(wb, v_int);
    size+= 4;
    break;
  case 'q':
    v_qua= va_arg(ap, uint64_t);
    if (wb) tssh_wb_put_uint64(wb, v_qua);
    size+= 8;
    break;
  case 'w':
    v_wor= va_arg(ap, int);
    if (wb) tssh_wb_put_uint16(wb, v_wor);
    size+= 2;
    break;
  case 'o':
    v_obj= va_arg(ap, tssh_obj_t*);
    if (wb) tssh_wb_put_obj(wb, v_obj);
    size+= v_obj->len;
    break;
  case 'B':
    v_int= va_arg(ap, uint32_t);
    v_bytp= va_arg(ap, uint8_t*);
    if (wb) 
    {
      memcpy(wb->data+wb->len, v_bytp, v_int);
      wb->len += v_int;
    }
    size+= v_int;
    break;
  case 'R':
    if (wb) tssh_wb_put_uint32(wb, ++T->ftp_reqid);
    size+= 4;
    break;
  case 'C':
    v_byt= va_arg(ap, int);
    if (wb) tssh_wb_put_uint8(wb, v_byt);
    size+= 1;
    break;
  }
  va_end(ap);  

  if (!wb)
  {
    wb= tssh_alloc_write_buffer(size);
    goto again;
  }
  tssh_put_uint32(wb->data, size-4);
  return wb;
}

static tssh_write_buffer_t* tssh_ftp_cmd_format
   (tssh_t *T,const char *fmt, ...)
{
  va_list ap;
  tssh_write_buffer_t *wb;
  va_start(ap, fmt);
  wb= tssh_ftp_cmd_formatp(T, fmt, ap);
  va_end(ap);
  return wb;
}


static void tssh_ftp_thread_msg(tssh_t *T, int cmd, void *data)
{
  coth_msg_put(T->ftp_ctx, cmd, data);
  coth_switch(T->main_ctx, T->ftp_ctx);
}

static int tssh_ftp_cmd(tssh_t *T, const char *fmt, ...)
{
  va_list ap;
  tssh_write_buffer_t *wb;
  tssh_channel_t *ch;

  ch= T->ftp_channel;
  if (!ch) return 1;

  va_start(ap, fmt);
  wb= tssh_ftp_cmd_formatp(T, fmt, ap);
  va_end(ap);

  if (!ch->pend_last) ch->pend_first= wb;
                 else ch->pend_last->next= wb;
  ch->pend_last= wb;

  tssh_forward_ftp_write(T);
  return 0;
}
