static void tssh_ftp_main(tssh_t *T)
{
  int cmd;
  void *data;
  tssh_ftop_t *O;
  int R;

  tssh_ftp_do_connect(T);

again:
  if (!T->ftp_channel) goto fail_hard;
  if (coth_msg_get(T->ftp_ctx, &cmd, &data)) goto check_oplist;

  // we have a message, for the ftp subthread from the main thread.
  // however, all messages except for a possible _DISCONNECT should
  // be handled inside the operation functions. whatever this message
  // is, it arrived at the incorrect time or is a genuine disconnect.
  // we therefore end this channel and inform the user.
  if (data) free(data);
  goto fail_hard;

check_oplist:
  O= T->ftop_first;
  if (!O) goto wait;

  T->ftop_current= O;
  T->ftop_first= O->next;
  if (!T->ftop_first) T->ftop_last= NULL;

  switch(O->cmd)
  {
  case TSSH_FTOP_LIST: R= tssh_ftp_do_list(T); break;
  case TSSH_FTOP_DOWNLOADMEM: R= tssh_ftp_do_downloadmem(T); break;
  case TSSH_FTOP_DOWNLOADFILE: R= tssh_ftp_do_downloadfile(T); break;
  case TSSH_FTOP_UPLOADFILE: R= tssh_ftp_do_uploadfile(T); break;
  case TSSH_FTOP_UPLOADMEM: R= tssh_ftp_do_uploadmem(T); break;
  case TSSH_FTOP_RMFILE: R= tssh_ftp_do_rmfile(T); break;
  case TSSH_FTOP_RMDIR: R= tssh_ftp_do_rmdir(T); break;
  case TSSH_FTOP_RENAME: R= tssh_ftp_do_rename(T); break;
  case TSSH_FTOP_MKDIR: R= tssh_ftp_do_mkdir(T); break;
  case TSSH_FTOP_QUERY: R= tssh_ftp_do_query(T); break;
  }
  switch(R)
  {
  case TSSH_FAIL:
    tssh_ftp_report_opfail(T, T->ftop_current);
    T->ftop_current= NULL;
    break;
  case TSSH_FAIL_HARD:
    goto fail_hard;
  }
  goto again;

wait:
  coth_switch(T->ftp_ctx, T->main_ctx);
  goto again;
// this mode of failure is a protocol failure, not just a file system
// failure. so, we don't have any hope of recovering from it.
fail_hard:
  tssh_ftp_die(T, TSSH_EVENT_FTP_DISCONNECTED);
}


