
static int tssh_msg_channel_open_confirmation(tssh_t *T, tssh_buffer_t *pb)
{
  uint8_t cmd;
  uint32_t local_number, remote_number, window_size, max_packet_size;
  tssh_channel_t *ch;
  if (tssh_xparse_byte("cmd", pb, &cmd) ||
      tssh_xparse_uint32("localid", pb, &local_number) ||
      tssh_xparse_uint32("remoteid", pb, &remote_number) ||
      tssh_xparse_uint32("windowsize", pb, &window_size) ||
      tssh_xparse_uint32("maxpacketsize", pb, &max_packet_size))
  {
    tssh_log(T,"bad channel open confirmation message\n");
    return 1;
  }
  ch= tssh_find_channel(T, local_number);
  if (!ch) return 1;
  ch->remote_number= remote_number;
  ch->max_packet_size= tssh_min_uint32(ch->max_packet_size, max_packet_size);
  ch->window_size_c2s= window_size;
  switch(ch->type)
  {
  case TSSH_CHANNEL_SHELL: 
    tssh_shell_request_pty(T, ch);
    break;
  case TSSH_CHANNEL_FTP:
    tssh_ftp_channel_open_success(T);
    break;
  }
  return 0;
}

static int tssh_msg_channel_open_failure(tssh_t *T, tssh_buffer_t *pb)
{
  tssh_event_t *E;
  uint8_t cmd;
  uint32_t local_number, reason_code; tssh_obj_t descr, lang;
  tssh_channel_t *ch;
  if (tssh_xparse_byte("cmd", pb, &cmd) ||
      tssh_xparse_uint32("localid",pb, &local_number) ||
      tssh_xparse_uint32("reason", pb, &reason_code) ||
      tssh_xparse_string("descr", pb, &descr) ||
      tssh_xparse_string("lang", pb, &lang))
  {
    tssh_log(T,"bad channel open failure message\n");
    return 1;
  }
  tssh_log(T, "channel-%d open failed: %S\n", local_number, &descr);
  ch= tssh_find_channel(T, local_number);
  if (!ch) return 0;
  E= NULL;
  switch(ch->type)
  {
  case TSSH_CHANNEL_SHELL: 
    E= tssh_event_new(TSSH_EVENT_SHELL_FAILURE);
    break;
  case TSSH_CHANNEL_FTP:
    tssh_ftp_channel_open_failure(T);
    break;
  }
  tssh_channel_delete(T,ch);
  if (E) tssh_report(T,E);
  return 0;
}

static int tssh_msg_channel_failure(tssh_t *T, tssh_buffer_t *pb)
{
  uint8_t cmd;
  uint32_t localid;
  tssh_channel_t *ch;
  if (tssh_xparse_byte("cmd", pb, &cmd) ||
      tssh_xparse_uint32("localid", pb, &localid))
  {
    tssh_log(T, "bad channel failure message\n");
    return 1;
  }
  ch= tssh_find_channel(T, localid);
  if (!ch) return 0;
  switch(ch->type)
  {
  case TSSH_CHANNEL_SHELL: return tssh_shell_channel_failure(T, ch);
  case TSSH_CHANNEL_FTP: tssh_ftp_channel_failure(T); return 0;
  }
  return 0;
}

static int tssh_msg_channel_success(tssh_t *T, tssh_buffer_t *pb)
{
  uint8_t cmd;
  uint32_t localid;
  tssh_channel_t *ch;
  if (tssh_xparse_byte("cmd", pb, &cmd) ||
      tssh_xparse_uint32("localid", pb, &localid))
  {
    tssh_log(T, "bad channel success message\n");
    return 1;
  }
  ch= tssh_find_channel(T, localid);
  if (!ch) return 0;
  switch(ch->type)
  {
  case TSSH_CHANNEL_SHELL: return tssh_shell_channel_success(T, ch);
  case TSSH_CHANNEL_FTP: tssh_ftp_channel_success(T); return 0;
  }
  return 0;
}


static int tssh_msg_channel_eof(tssh_t *T, tssh_buffer_t *pb)
{
  uint8_t cmd; uint32_t localid;
  tssh_channel_t *ch;
  if (tssh_parse_byte(pb, &cmd) ||
      tssh_xparse_uint32("localid", pb, &localid))
  {
    tssh_log(T, "bad channel_eof message\n");
    return 1;
  }
  ch= tssh_find_channel(T, localid);
  if (!ch) return 0;
  switch(ch->type)
  {
  case TSSH_CHANNEL_PORT_FORWARD:
   printf("port forward EOF\n");
    tssh_port_forward_eof(T, ch);
    break;
  }
  return 0;
}
