#include <signal.h>
#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include "allcom.c"
#include "working.c"

static int consolefd()
{
  int fd;
  struct sockaddr_in addr;
  fd= socket(AF_INET, SOCK_STREAM,0);
  addr.sin_family= AF_INET;
  addr.sin_port= htons(4321);
  inet_pton(AF_INET, "127.0.0.1", &addr.sin_addr);
  if (connect(fd, (struct sockaddr*) &addr, sizeof(addr)))
   exit(fprintf(stderr,"can not connect to console: %s\n", strerror(errno)));
  return fd;
}


static tssh_channel_t *shell_ch, *sock_ch;

void read_console(tssh_t *T,int cfd)
{
  uint8_t buf[1];
  int R;
  R= read(cfd, buf, sizeof(buf));
  if (R<=0) return ;
  if (!shell_ch) return ;
  tssh_channel_write(shell_ch, buf, R);
}

void csend(int fd, uint8_t *buf, int len)
{
  int R;
  while(len)
  {
    R= send(fd, buf, len, MSG_NOSIGNAL);
    if (R<0) { perror("send"); return ; }
    buf+= R;
    len-= R;
  }
}

static int sote_recv;
static int sote_state;
static int sote_len;
static int sote_cmd;
static uint32_t sote_port;

static void sock_send_data()
{
#define WSIZE 20480
  int len;
  tssh_write_buffer_t *B;
  if (!sote_len) return ;
  len= rand() % WSIZE;
  if (len>sote_len) len= sote_len;
  B= tssh_alloc_write_buffer(len);
  B->len= len;
  tssh_channel_write_buffer(sock_ch, B);
  sote_len-= len;
}

static void process_socket_data(tssh_buffer_t *buffer)
{
  int cp;
again:
  if (!buffer->len) return ;
  switch(sote_state) 
  {
  case 0: goto init; 
  case 1: goto getlen;
  case 2: goto consume;
  }
  return ;
init:
  sote_cmd= buffer->data[0];
  buffer->data++; 
  buffer->len--;
  sote_state= 1;
  sote_len= 0;
  goto again;
getlen:
  if (buffer->data[0]=='.')
  {
    buffer->data++; buffer->len--; sote_state= 2; 
    if (sote_cmd=='R') goto send_data;
    goto again;
  }
  sote_len= sote_len*10 + buffer->data[0] - '0';
  buffer->data++; buffer->len--; goto again;
consume:
  if (sote_cmd=='M') goto consume_M;

  cp= buffer->len;
  if (cp>sote_len) cp= sote_len;
  if (sote_recv==0)
  {
     if (buffer->data[0]!='<') printf("begin-marker not present\n");
  }
  if (cp==sote_len)
  {
    if (buffer->data[cp-1]!='>') printf("end-marker not present\n");
  }
  buffer->data+= cp;
  buffer->len-= cp;
  sote_len-= cp;
  sote_recv+= cp;
  if (sote_len==0) 
  {
    sote_state= 0;
    printf("%d bytes received\n", sote_recv);
    sote_recv= 0;
  }
  printf("progress %d bytes\n", sote_recv);
  goto again;

consume_M:
  cp= buffer->len;
  if (cp>sote_len) cp= sote_len;
  fwrite(buffer->data, 1, cp, stdout);
  buffer->data+= cp;
  buffer->len-= cp;
  sote_len-= cp;
  if (sote_len==0) sote_state= 0;
  goto again;
send_data:
  sock_send_data();
}

void read_event(tssh_t *T,int cfd)
{
  tssh_event_t *E;
  tssh_buffer_t *buffer;
  char buf[400];

  E= tssh_get_event(T);
  switch(E->type)
  {
  case TSSH_EVENT_PORT_FORWARD_CONNECT:
    printf("incoming connection to peer %d@%s from %d@%s\n",
             E->connected_port, E->connected_address,
             E->originator_port, E->originator_address);
    sock_ch= E->channel;
    break;
  case TSSH_EVENT_PORT_FORWARD_SUCCESS:
    printf("port forward request succeeded: peer port= %d\n",E->port);
    sote_port= E->port;
    tssh_shell(T, NULL); 
    break;
  case TSSH_EVENT_LOG: 
    printf("L>%s", E->D.str); break;
  case TSSH_EVENT_CONNECT_BANNER: printf("B>%s", E->D.str); break;
  case TSSH_EVENT_PASSWORD_BANNER: printf("P>%s", E->D.str); break;
  case TSSH_EVENT_CONNECT_SUCCESS: 
     printf("Connection OK. Trying port forwarding.\n");
     tssh_request_port_forward(T, 0);
     break;
  case TSSH_EVENT_CONNECT_FAILURE: printf("Connection Failed\n"); break;
  case TSSH_EVENT_SHELL_OK: 
     printf("Shell OK\n"); 
     shell_ch= E->channel;

     sprintf(buf,"SOTEPORT=%d\n", sote_port);
     tssh_channel_write(shell_ch, buf, strlen(buf));
     sprintf(buf,"export SOTEPORT\n");
     tssh_channel_write(shell_ch, buf, strlen(buf));
     break;
  case TSSH_EVENT_SHELL_FAILURE: printf("Shell failed\n"); break;
  case TSSH_EVENT_CHANNEL_EXTENDED_DATA:
  case TSSH_EVENT_CHANNEL_DATA:
     if (E->channel==shell_ch)
     {
       buffer= E->buffer;
       csend(cfd, buffer->data, buffer->len);
     } 
     else
     {
       process_socket_data(E->buffer);
     }
     break;
  case TSSH_EVENT_CHANNEL_WRITE_OK:
     if (E->channel==sock_ch) { sock_send_data(); }
     // printf("Write OK\n");
     break;
  case TSSH_EVENT_CHANNEL_CLOSED:
     printf("Channel closed: ");
     if (E->channel==shell_ch) { printf("Shell\n"); shell_ch= NULL; }
     else if (E->channel==sock_ch) { printf("Socket\n"); sock_ch= NULL;}
     else { printf("Unknown\n"); }
     break;
  }
}

int main()
{
  tssh_t *T;
  char *cmd;
  fd_set rset, xset;
  int sfd, cfd, maxfd;
  int R;

  T= tssh_init(NULL, "localhost", -1, "test", "hunter2");

  cmd= "ls -l /usr/bin\n";
  sfd= tssh_event_fd(T);
  cfd= consolefd();
  maxfd= cfd>sfd ? cfd: sfd;

again:
  FD_ZERO(&rset); FD_ZERO(&xset);
  FD_SET(sfd, &rset);
  FD_SET(cfd, &rset);
  FD_SET(sfd, &xset);
  R= select(maxfd+1, &rset, NULL, &xset, NULL);
  if (R<0 && errno==EINTR) goto again;
  if (FD_ISSET(sfd, &xset)) goto done;
  if (FD_ISSET(sfd, &rset)) read_event(T,cfd);
  if (FD_ISSET(cfd, &rset)) read_console(T,cfd);
  goto again;

done:
  close(T->fd);
  return 0;
}
