#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "../components/0021-aes256.c"

char *NISTSP_800_38A_AES256CTR_Encrypt[]= {
  "603deb1015ca71be2b73aef0857d7781 1f352c073b6108d72d9810a30914dff4", // key
  "f0f1f2f3f4f5f6f7f8f9fafbfcfdfeff", // iv

  "f0f1f2f3f4f5f6f7f8f9fafbfcfdfeff", // input to Cipher (set by xcrypt)
  "0bdf7df1591716335e9a8b15c860c502", // output from Cipher 
  "6bc1bee22e409f96e93d7e117393172a", // plaintext
  "601ec313775789a5b7a7f504bbf3d228", // ciphertext

  "f0f1f2f3f4f5f6f7f8f9fafbfcfdff00",
  "5a6e699d536119065433863c8f657b94",
  "ae2d8a571e03ac9c9eb76fac45af8e51",
  "f443e3ca4d62b59aca84e990cacaf5c5",

  "f0f1f2f3f4f5f6f7f8f9fafbfcfdff01",
  "1bc12c9c01610d5d0d8bd6a3378eca62",
  "30c81c46a35ce411e5fbc1191a0a52ef",
  "2b0930daa23de94ce87017ba2d84988d",

  "f0f1f2f3f4f5f6f7f8f9fafbfcfdff02",
  "2956e1c8693536b1bee99c73a31576b6",
  "f69f2445df4f9b17ad2b417be66c3710",
  "dfc9c58db67aada613c2dd08457941a6",
  NULL
};

char *NISTSP_800_38A_AES256CTR_Decrypt[]= {
  "603deb1015ca71be2b73aef0857d7781 1f352c073b6108d72d9810a30914dff4",
  "f0f1f2f3f4f5f6f7f8f9fafbfcfdfeff",

  "f0f1f2f3f4f5f6f7f8f9fafbfcfdfeff",
  "0bdf7df1591716335e9a8b15c860c502",
  "601ec313775789a5b7a7f504bbf3d228",
  "6bc1bee22e409f96e93d7e117393172a",

  "f0f1f2f3f4f5f6f7f8f9fafbfcfdff00",
  "5a6e699d536119065433863c8f657b94",
  "f443e3ca4d62b59aca84e990cacaf5c5",
  "ae2d8a571e03ac9c9eb76fac45af8e51",

  "f0f1f2f3f4f5f6f7f8f9fafbfcfdff01",
  "1bc12c9c01610d5d0d8bd6a3378eca62",
  "2b0930daa23de94ce87017ba2d84988d",
  "30c81c46a35ce411e5fbc1191a0a52ef",
 
  "f0f1f2f3f4f5f6f7f8f9fafbfcfdff02",
  "2956e1c8693536b1bee99c73a31576b6",
  "dfc9c58db67aada613c2dd08457941a6",
  "f69f2445df4f9b17ad2b417be66c3710",
  NULL
};

static int hexv(char c)
{
  if (c>='0' && c<='9') return c-'0';
  if (c>='a' && c<='f') return c-'a'+10;
  if (c>='A' && c<='F') return c-'A'+10;
  return -1;
}

static void hex2bin(char *inp, uint8_t *outp, int explen)
{
  int ndigits, i, shift, advance, k;
  char c;
  int len;
  ndigits= 0;
  for(i=0;inp[i];i++) if (hexv(inp[i])>=0) ndigits++;
  if (ndigits%2) { shift= 0; advance= 1; }
            else { shift= 4; advance= 0; }
  len= (ndigits+1)/2;
  if (len!=explen)
  {
    printf("expected length for string was %d, found %d\n", 
                    explen, len);
    exit(1);
  }
  memset(outp, 0, len);
  while((c=*(inp++)))
  {
    k= hexv(c);
    if (k<0) continue;
    *outp= *outp | (k<<shift);
    outp+= advance;
    shift= 4-shift;
    advance= 1-advance;
  }
}

static void dump_hex(char *msg, uint8_t *d, int l)
{
  int i;
  printf("%s", msg);
  for(i=0;i<l;i++) printf("%02x ", d[i]);
  printf("\n");
}

static void test_NISTSP_800_38A_AES256CTR(char **tv)
{
  uint8_t *key, *iv, *pt, *ct;
  int bn;
  struct AES256_ctx *ctx;

   // variables below are allocated using malloc() so that
   // valgrind keeps track and recognizes overruns.
  ctx= malloc(sizeof(*ctx)); 
  key= malloc(32);
  iv= malloc(16);
  pt= malloc(16);
  ct= malloc(16);

  hex2bin( tv[0], key, 32);
  hex2bin( tv[1], iv, 16);
  tv+= 2;

  AES256CTR_init(ctx, key, iv);
  bn= 0;
  while(tv[0])
  {
    printf("block %d\n", bn);
    hex2bin(tv[2], pt, 16);
    hex2bin(tv[3], ct, 16);
    dump_hex("PT: ", pt,16);
    dump_hex("CT: ", ct,16);
    AES256CTR_xcrypt(ctx, pt, 16);
    if (memcmp(pt,ct,16))
    {
      printf("test failed at block %d\n", bn);
      dump_hex("XT: ", pt, 16);
      exit(1);
    }
    tv+= 4;
    bn++;
  }
}

int main()
{
  printf("testing NIST SP 800-38A vectors\n");
  printf("vector 1\n");
  test_NISTSP_800_38A_AES256CTR(NISTSP_800_38A_AES256CTR_Encrypt);
  printf("vector 2\n");
  test_NISTSP_800_38A_AES256CTR(NISTSP_800_38A_AES256CTR_Decrypt);
  printf("OK\n");
  return 0;
}
