#include <unistd.h> 
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <termios.h>
#include <string.h>
#include <sys/socket.h>
#include <arpa/inet.h>

static void die(char *msg)
{
  if (msg)
  {
    if (msg[0]=='!') perror(msg+1);
    else fprintf(stderr,"%s", msg);
    fprintf(stderr,"\n");
  }
  exit(1);
}

static void console_raw_mode(int ONoff)
{
  static struct termios orig_termios;
  struct termios raw;
  if (ONoff)
  {
    tcgetattr(STDIN_FILENO, &orig_termios);
    raw= orig_termios;
    raw.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
    raw.c_oflag &= ~(OPOST);
    raw.c_cflag |= (CS8);
    raw.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
    tcsetattr(STDIN_FILENO, TCSAFLUSH, &raw);
  }
  else
  {
    tcsetattr(STDIN_FILENO, TCSAFLUSH, &orig_termios);
  }
}
                                                     

int servsock(int port)
{
  int fd;
  struct sockaddr_in addr;
  int yes;
  
  fd= socket(AF_INET, SOCK_STREAM, 0);
  if (fd<0) die("!socket");

  setsockopt(fd, SOL_SOCKET, SO_REUSEADDR , &yes, sizeof(yes));

  addr.sin_family= AF_INET;
  addr.sin_addr.s_addr= INADDR_ANY;
  addr.sin_port= htons(port);

  if (bind(fd,(struct sockaddr *) &addr, sizeof(addr))) die("!bind");

  if (listen(fd, 5)) die("!listen");
  return fd;
}

int getsock(int sfd)
{
  int fd;
  printf("\r\nwaiting for a connection\r\n");
  fd= accept(sfd, NULL, NULL);
  if (fd<0) die("!accept");
  printf("connected\r\n");
  return fd;
}
  
void xfer(int src, int dst)
{
  char buf[1024];
  int R;
  char *P; int L;
  
  R= read(src, buf, sizeof(buf));
  if (R<=0) exit(fprintf(stderr,"xfer %d -> %d : read: %s\n", 
                              src,dst, strerror(errno)));
  for(L=R,P=buf;L;L-=R,P+=R)
  {
    R= write(dst, P, L);
    if (R<=0) exit(fprintf(stderr,"xfer %d -> %d : write: %s\n", 
                              src,dst, strerror(errno)));
  }
}

int main(int argc , char *argv[])
{
  int sfd, fd;
  int R;
  fd_set rset, xset;

  console_raw_mode(1);

  //sfd= servsock(atoi(argv[1]));
  sfd= servsock(4321);
done:
  fd= getsock(sfd);
again:
  FD_ZERO(&rset); FD_ZERO(&xset);
  FD_SET(fd, &rset);
  FD_SET(0, &rset);
  FD_SET(fd, &xset);
  R= select(fd+1, &rset, NULL, &xset, NULL);
  if (R<0 && errno==EINTR) goto again;
  if (R<0) die("!select");
  if (R==0) goto again;
  if (FD_ISSET(fd, &xset)) goto done;
  if (FD_ISSET(fd, &rset)) xfer(fd, 1);
  if (FD_ISSET(0, &rset)) xfer(0, fd);
  goto again;     
  return 0;
}
