#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <string.h>
#include <errno.h>

static void die(const char *msg)
{
  char *es;
  if (msg[0]=='!') { es= strerror(errno); msg++; }
              else es= NULL;
  fprintf(stderr,"%s",msg);
  if (es) fprintf(stderr,": %s", es);
  fprintf(stderr,"\n");
  exit(1);
}

static int get_len(char *str)
{
  int len;
  len= atoi(str);
  if (strchr(str,'k') || strchr(str,'K')) len *= 1024;
  else if (strchr(str,'m') || strchr(str,'M')) len *= 1024*1024;
  else if (strchr(str,'g') || strchr(str,'G')) len *= 1024*1024*1024;
  return len;
}

static int send_str(int fd, char *str)
{
  int R;
  while(*str)
  {
    R= send(fd, str, strlen(str), MSG_NOSIGNAL);
    if (R==0) return 1;
    if (R<0 && errno==EINTR) continue;
    if (R<0) die("!send");
    str+= R;
  }
  return 0;
}

int parse_cmd(char **Rs,char *cmd)
{
  char *s;
  int R;
  s= *Rs;
  R= 0;

  if (!strncmp(s, cmd, strlen(cmd))) { R= 1; s+= strlen(cmd)+1; }

  *Rs= s;
  return R;
}

void send_bytes(int fd, int len)
{
  char buf[4096];
  int R,cp;
  int first_packet;
  sprintf(buf,"S%d.", len);
  if (send_str(fd, buf)) goto fail;
  first_packet= 1;
  while(len)
  {
    cp= sizeof(buf);
    if (cp>len) cp= len;
    if (first_packet) buf[0]='<';
    if (cp==len) buf[cp-1]= '>';
    R= send(fd, buf, cp, MSG_NOSIGNAL);
    if (R<=0) goto fail;
    first_packet= 0;
    len -= R;
  }
  return ;
fail:
  exit(fprintf(stderr,"send_bytes: %s\n", strerror(errno)));
}

void get_bytes(int fd, int len)
{
  char buf[40960];
  int R,cp;
  sprintf(buf, "R%d.", len);
  if (send_str(fd, buf)) goto fail;

  while(len)
  {
    cp= sizeof(buf);
    if (cp>len) cp= len;
    R= recv(fd, buf, cp, 0);
    if (R<=0) goto fail;
    len -= R;
  }

  return ;
fail:
  exit(fprintf(stderr,"get_bytes: %s\n", strerror(errno)));
}

int main()
{
  int fd;
  fd_set rset, xset;
  int R;
  struct sockaddr_in addr;
  char *port;
  char cbuf[1024];
  char sbuf[1024];
  char *str;
  int len;

  port= getenv("SOTEPORT");
  if (!port) die("you must set SOTEPORT first\n");
  fd= socket(AF_INET, SOCK_STREAM, 0);
  if (fd<0) die("!socket");
  memset(&addr, 0, sizeof(addr));
  addr.sin_family= AF_INET;
  inet_pton(AF_INET, "127.0.0.1", &addr.sin_addr);
  addr.sin_port= htons(atoi(port));
  if (connect(fd, (struct sockaddr*) &addr, sizeof(addr))) die("!connect");

again:
  FD_ZERO(&rset);
  FD_ZERO(&xset);
  FD_SET(0, &rset);
  FD_SET(fd, &rset);
  FD_SET(fd, &xset);

  R= select(fd+1, &rset, NULL, &xset, NULL);
  if (R<0 && errno==EINTR) goto again;
  if (R<0) die("!select");
  if (R==0) goto again;
  if (FD_ISSET(fd, &xset)) goto fd_closed;
  if (FD_ISSET(fd, &rset)) goto fd_read;
  if (FD_ISSET(0, &rset)) goto console_read;
  goto again;
fd_closed:
  printf("Closed by remote end.\n");
  close(fd);
  return 0;
console_read:
  R= read(0, cbuf, sizeof(cbuf));
  if (R<=0) goto again; // cancelled or EOFd or something
  cbuf[R]= 0;
  str= cbuf;
  if (parse_cmd(&str, "m")) 
  {
    char buf[20];
    sprintf(buf,"M%d.", strlen(str));
    if (send_str(fd, buf)) goto fd_closed; 
    if (send_str(fd, str)) goto fd_closed;
    goto again;
  }
  if (parse_cmd(&str,"s"))
  {
    len= get_len(str);
    printf("Sending %d bytes\n", len);
    send_bytes( fd, len );
    printf("Sent all\n");
    goto again;
  }
  if (parse_cmd(&str,"r"))
  {
    len= get_len(str);
    printf("Will receive %d bytes\n", len);
    get_bytes(fd, len);
    printf("Received all\n");
    goto again;
  }
  if (parse_cmd(&str,"q"))
  {
    close(fd); 
    exit(0);
  }
  goto again;
fd_read:
  R= recv(fd, sbuf, sizeof(sbuf), 0);
  if (R==0) goto fd_closed;
  if (R<0 && errno==EINTR) die("!recv");
  printf("idata: ");
  fwrite(sbuf, 1, R, stdout);
  printf("\n");
  goto again;
} 

