

static void tssh_main_thread(tssh_t *T)
{
  int cmd; void *ap1,*ap2,*ap3; uint32_t ai1;
  tssh_obj_t *obj;
  tssh_obj_t strobj;
  tssh_buffer_t *pb;
  uint32_t localid;
  uint8_t msg;
  tssh_channel_t *ch;
  uint8_t Dbool, Dbyte; uint32_t Du32; tssh_obj_t Dstr; 

  if (tssh_connect(T)) goto fail_connect;
  if (tssh_send_self_identification(T, "SSH-2.0-tinyssh_1.0 Quit complaining.")
      || tssh_get_remote_identification(T,&obj)) goto fail;

  tssh_put_thread_msg(&T->reader_queue,0, TSSH_CMD_START_READER, obj,NULL,0,0);
  tssh_send_kexinit(T);

  pb= NULL;
again:
  if (pb) { free(pb); pb= NULL; }
  tssh_get_thread_msg(&T->main_queue, &cmd, &ap1, &ap2, &ai1, &ap3);
  switch(cmd)
  {
  case TSSH_CMD_QUIT: goto again;
  case TSSH_CMD_NEWPACKET: goto handle_new_packet;
  case TSSH_CMD_ERROR: goto have_error;
  case TSSH_CMD_CHANNEL_WRITE: 
     if (tssh_forward_channel_write(T, ap1)) goto fail;
     break;
  case TSSH_CMD_CHANNEL_WRITE_OK:
     if (tssh_channel_write_ok(T, ap1, ai1)) goto fail;
     break;
  case TSSH_CMD_SHELL:
     tssh_shell_start(T, ap1);
     break;
  case TSSH_CMD_PORT_FORWARD:
     tssh_send_port_forward_request(T, ai1);
     break;
  case TSSH_CMD_FTP_CONNECT: tssh_ftp_start(T); break;
  case TSSH_CMD_FTP_DISCONNECT: 
     if (T->ftp_ctx)
        tssh_ftp_thread_msg(T, TSSH_CMD_FTP_DISCONNECT, NULL);
     else
        tssh_report_simple(T, TSSH_EVENT_FTP_DISCONNECTED);
     break;
  case TSSH_CMD_DEALLOC_COTH: coth_free(ap1); break;
  case TSSH_CMD_QUEUE_FTOP: 
     if (T->ftp_ctx)
     {
       if (T->ftop_last) T->ftop_last->next= ap1;
                    else T->ftop_first= ap1;
       T->ftop_last= ap1;
       coth_switch(T->main_ctx, T->ftp_ctx);
     }
     else
       tssh_ftp_report_opfail(T, ap1);
     break;   
  }
  goto again;

handle_new_packet:
  pb= ap1;
#ifndef TSSH_NO_DEBUG
  printf("CMD= "); tssh_print_cmd_name(pb->data[0]); 
#endif
  switch(pb->data[0])
  {
  case SSH_MSG_DISCONNECT: 
  case SSH_MSG_IGNORE: goto again;
  case SSH_MSG_UNIMPLEMENTED:
  case SSH_MSG_DEBUG: goto again;
  case SSH_MSG_SERVICE_REQUEST: goto bad_side;
  case SSH_MSG_SERVICE_ACCEPT: 
    if (tssh_msg_service_accept(T,pb)) goto protocol_error;
    goto again;
  case SSH_MSG_EXT_INFO:
    goto unhandled;
  case SSH_MSG_KEXINIT:
    if (tssh_msg_kexinit(T,pb)) goto protocol_error;
    goto again;
  case SSH_MSG_NEWKEYS:
    if (tssh_msg_newkeys(T, pb)) goto protocol_error;
    goto again;
  case SSH_MSG_KEXDH_INIT: goto bad_side;
  case SSH_MSG_KEXDH_REPLY:
    if (tssh_msg_kexdh_reply(T,pb)) goto protocol_error;
    goto again;
  case SSH_MSG_USERAUTH_REQUEST: goto bad_side;
  case SSH_MSG_USERAUTH_PASSWD_CHANGEREQ:
  case SSH_MSG_USERAUTH_FAILURE:
    tssh_report_simple(T, TSSH_EVENT_CONNECT_FAILURE);
    goto again;
  case SSH_MSG_USERAUTH_SUCCESS:
    tssh_report_simple(T, TSSH_EVENT_CONNECT_SUCCESS);
    T->auth= 1;
    T->auth_state= 0;
    goto again;
  case SSH_MSG_USERAUTH_BANNER:
    tssh_msg_userauth_banner(T, pb);
    goto again;
  case SSH_MSG_GLOBAL_REQUEST: break;
  case SSH_MSG_REQUEST_SUCCESS:
    if (tssh_msg_global_request_success(T, pb)) goto fail;
    goto again;
  case SSH_MSG_REQUEST_FAILURE:
    if (tssh_msg_global_request_failure(T, pb)) goto fail;
    goto again;
  case SSH_MSG_CHANNEL_OPEN: 
    if (tssh_msg_channel_open(T, pb)) goto fail;
    break;
  case SSH_MSG_CHANNEL_OPEN_CONFIRMATION:
    if (tssh_msg_channel_open_confirmation(T, pb)) goto fail;
    break;
  case SSH_MSG_CHANNEL_OPEN_FAILURE:
    if (tssh_msg_channel_open_failure(T,pb)) goto fail;
    break;
  case SSH_MSG_CHANNEL_WINDOW_ADJUST:
    if (tssh_msg_window_adjust(T, pb)) goto fail;
    break;
  case SSH_MSG_CHANNEL_DATA:
    if (tssh_msg_channel_data(T, &pb)) goto fail;
    break;
  case SSH_MSG_CHANNEL_EXTENDED_DATA:
    if (tssh_msg_channel_extended_data(T, &pb)) goto fail;
    break;
  case SSH_MSG_CHANNEL_EOF:
    if (tssh_msg_channel_eof(T, pb)) goto fail;
    break;
  case SSH_MSG_CHANNEL_CLOSE:
    if (tssh_msg_channel_close(T, pb)) goto fail;
    break;
  case SSH_MSG_CHANNEL_REQUEST: 
    if (tssh_parse_byte(pb, &Dbyte) ||
        tssh_xparse_uint32("localid", pb, &Du32) ||
        tssh_xparse_string("request", pb, &Dstr) ||
        tssh_xparse_bool("wantrep", pb, &Dbool)) goto fail;
    break;
  case SSH_MSG_CHANNEL_SUCCESS:
    if (tssh_msg_channel_success(T,pb)) goto fail;
    break;
  case SSH_MSG_CHANNEL_FAILURE:
    if (tssh_msg_channel_failure(T,pb)) goto fail;
    break;
  }
  goto again;
fail_connect:
  return ;
protocol_error:
  return ;
bad_side:
  return ;
have_error:
  return ;
unhandled:
  return ;
fail:
  return ;
}

static void* tssh_main_thread_entry(void *arg)
{
  tssh_main_thread(arg);
  return NULL;
}

tssh_t *tssh_init(char *alias, char *host, int port, char *user, char *passwd)
{
  tssh_t *T;
  int eno;

  T= calloc(1,sizeof(*T));

  if (socketpair(AF_UNIX, SOCK_STREAM, 0, T->iface_fds))
  {
    eno= errno;
    free(T);
    errno= eno;
    return NULL;
  }

  T->host= strdup(host);
  T->alias= alias ? strdup(alias) : NULL;
  T->port= port < 0 ? 22 : port;
  T->user= strdup(user);
  T->password= strdup(passwd);

  T->avail_channel_no= 1;

  tssh_queue_init(&T->writer_queue);
  tssh_queue_init(&T->reader_queue);
  tssh_queue_init(&T->main_queue);
  tssh_queue_init(&T->iface_queue);

  tssh_array_init(&T->a_fwreq, (void**) &T->fwreq, &T->nfwreq, 
                    sizeof(uint32_t), 10);

  pthread_create(&T->writer_thread, NULL, tssh_writer_thread_entry, T);
  pthread_create(&T->reader_thread, NULL, tssh_reader_thread_entry, T);
  pthread_create(&T->main_thread, NULL, tssh_main_thread_entry, T);
  T->main_ctx= coth_main();

  return T;
}

